import numpy as np
import moderngl as mgl

from ..core import System


class Field3DBuffer:
    def __init__(self, mglctx: mgl.Context = None, system: System = None):
        self._ctx = mglctx
        self._system = system
        self.vbo = self._ctx.buffer(reserve=12 * self._system.number_of_spins)

    def upload(self, state: np.ndarray):
        assert isinstance(state, np.ndarray)
        self.vbo.write(state.reshape((-1, 3)).astype("f4").tobytes())


class Field1DBuffer:
    def __init__(self, mglctx: mgl.Context = None, system: System = None):
        self._ctx = mglctx
        self._system = system
        self.vbo = self._ctx.buffer(reserve=4 * self._system.number_of_spins)

    def upload(self, state: np.ndarray):
        assert isinstance(state, np.ndarray)
        self.vbo.write(state.flatten().astype("f4").tobytes())


class Float3DBuffer:
    def __init__(self, mglctx: mgl.Context = None, count: int = None):
        self._ctx = mglctx
        self._count = count
        self.vbo = self._ctx.buffer(reserve=12 * self._count)

    def upload(self, data: np.ndarray):
        assert isinstance(data, np.ndarray)
        data = data.reshape((-1, 3))
        assert data.shape[0] == self._count
        self.vbo.write(data.astype("f4").tobytes())

    def download(self):
        return np.frombuffer(self.vbo.read(), dtype=np.float32).reshape(
            (-1, 3)
        )  # , count=self._count)


class IntBuffer:
    def __init__(self, mglctx: mgl.Context = None, count=None):
        self._ctx = mglctx
        self._count = count
        self.ibo = self._ctx.buffer(reserve=4 * self._count)

    @property
    def count(self):
        return self._count

    def upload(self, data: np.ndarray):
        assert isinstance(data, np.ndarray)
        data = data.flatten()
        assert data.shape == (self._count,)
        self.ibo.write(data.astype("i4").tobytes())

    def download(self):
        return np.frombuffer(self.ibo.read(), dtype=np.int32)  # , count=self._count)



from .image import ImageContainer
from .camera import Camera
from .scene import (
    LayerScene,
    CubeScene,
    FlowScene,
    AstraScene,
    LevelScene,
    ChargeScene,
    VectorScene,
    IsolinesScene,
    SphereDotScene,
    LayerAndSurfaceScene,
    LayerAndIsolinesScene,
    SCENES,
    print_scenes,
)
from .sugar import render_layer, render_isosurface, render_isolines, render_prepare

__all__ = [
    "ImageContainer",
    "Camera",
    "LayerScene",
    "CubeScene",
    "FlowScene",
    "AstraScene",
    "LevelScene",
    "ChargeScene",
    "VectorScene",
    "IsolinesScene",
    "SphereDotScene",
    "LayerAndSurfaceScene",
    "LayerAndIsolinesScene",
    "SCENES",
    "print_scenes",
    "render_layer",
    "render_isosurface",
    "render_isolines",
    "render_prepare",
]



import moderngl as mgl
import numpy as np

from ..core import System


class Render:
    """
    Abstract class to run ModernGL based render.
    """

    VERTEX = """
                #version 330
                in vec3 in_spin;
                void main() {
                    gl_Position = vec4(in_spin, 1.0);
                }
            """
    GEOMETRY = None
    FRAGMENT = """
                #version 330
                out vec4 f_color;
                void main() {
                    f_color = vec4(1.0,1.0,1.0,1.0);
                }
            """

    def __init__(self, mglctx: mgl.Context = None):
        self._ctx = mglctx

        self._prog = self._ctx.program(
            vertex_shader=self.VERTEX,
            geometry_shader=self.GEOMETRY,
            fragment_shader=self.FRAGMENT,
        )

        self.populate_uniforms()

    def populate_uniforms(self):
        for u in self._prog:
            obj = self._prog[u]
            if isinstance(obj, mgl.Uniform):
                self.__dict__[u] = obj

    def render(self, **kwargs):
        self._ctx.clear(0.60, 0.60, 0.60)

    def list_properties(self):
        return []


####################################################################################


class StateRender(Render):
    """
    Abstract class for magnetization render.
    """

    def __init__(
        self,
        mglctx: mgl.Context = None,
        system: System = None,
        stateBuffer: np.ndarray = None,
    ):
        super().__init__(mglctx=mglctx)
        self._system = system
        self._state = stateBuffer

        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self._state.vbo, "3f", "in_spin"),
            ],
        )



#####################################################################################


class Property:
    def __init__(self, owner: object, field: str, title="noname", reset=False):
        self._owner = owner
        self._field = field
        self.title = title
        self.reset = reset

    def get(self):
        return getattr(self._owner, self._field)

    def let(self, value):
        setattr(self._owner, self._field, value)

    def __str__(self):
        resetable = "R " if self.reset else ""
        return f"{resetable}'{self.title}'"

    def __repr__(self):
        return str(self)


######################################################################################


class NumericProperty(Property):
    def __init__(self, *vargs, min=0.0, max=1.0, count=1, **kwargs):
        super().__init__(*vargs, **kwargs)
        self.min = min
        self.max = max
        self.count = count

    def __str__(self):
        prefix = super().__str__()
        return f"{prefix} = {self.get()}. Numeric {self.min} .. {self.max}"


######################################################################################


class BooleanProperty(Property):
    def __init__(self, *vargs, **kwargs):
        super().__init__(*vargs, **kwargs)

    def __str__(self):
        prefix = super().__str__()
        return f"{prefix} = {self.get()}"


######################################################################################


class ListProperty(Property):
    def __init__(self, *vargs, values=[], **kwargs):
        assert len(values) > 0
        super().__init__(*vargs, **kwargs)
        self.values = values

    def get(self):
        value = getattr(self._owner, self._field)
        assert value in self.values
        return value

    def let(self, value):
        assert value in self.values
        setattr(self._owner, self._field, value)

    def get_index(self, value=None):
        if value is None:
            value = self.get()
        return next((n for n, v in enumerate(self.values) if v == value))

    def let_index(self, index):
        self.let(self.values[index])

    def __str__(self):
        prefix = super().__str__()
        choices = ", ".join(f"{n}: `{v}`" for n, v in enumerate(self.values))
        return f"{prefix} = {self.get()}. One of {choices}"


######################################################################################


class PropertiesCollection:
    def __init__(self, props:list[Property]):
        self._props = props
        self._extract_names()

    def _extract_names(self):
        self._names = { p.title:n for n, p in enumerate(self._props) }

    def _resolve(self, key):
        if isinstance(key, str):
            if key not in self._names:
                raise KeyError(f"Property `{key}` does not exist.")
            return self._names[key]
        if isinstance(key, int):
            N = len(self._props)
            if not (0<=key<N):
                raise KeyError(f"Property index `{key}` does not belong to 0..{N}.")
            return key
        raise KeyError(f"Key type `{type(key)}` is not int or str.")

    def prop(self, key) -> Property:
        return self._props[self._resolve[key]]

    def __getitem__(self, key):
        return self._props[self._resolve(key)].get()

    def __setitem__(self, key, value):
        self._props[self._resolve(key)].let(value)

    def __str__(self):
        return ', '.join(f"{n}:'{k}'" for k, n in self._names)

    def __repr__(self):
        return str(self)


######################################################################################



import numpy as np

from ..core import System


def snap_vector(vec, tol=0.4):
    target = np.round(2 * vec) / 2
    target /= np.linalg.norm(target)
    return target if np.linalg.norm(target - vec) < tol else vec


class Camera:
    def __init__(
        self,
        look_at=[0, 0, 0],
        look_direction=[0, 0, 1],
        look_up=[1, 0, 0],
        scale=1.0,
        perspective=True,
    ):
        """
        Initialize camera position.
        """
        self.depth = 1000
        self.MAG = 1.5
        self.MAGA = (self.MAG + 1.0 / self.MAG) / 2
        self.perspective_is_on = perspective
        self.look_at = np.asarray(look_at, dtype=np.float32)
        self.look_direction = np.asarray(look_direction, dtype=np.float32)
        self.look_up = np.asarray(look_up, dtype=np.float32)
        self.look_scale = scale
        self.normalize()
        # self.set_predefined_direction(2)

    @classmethod
    def from_angles(self, theta, phi, alpha, scale, perspective, look_at):
        camera = Camera(scale=scale, perspective=perspective, look_at=look_at)
        camera.set_orientation_by_angles(theta=theta, phi=phi, alpha=alpha)
        return camera

    def __str__(self):
        return self.__repr__()
        # return "along {} up {} at {} scale {} per. {}".format(self.look_direction, self.look_up, self.look_at, self.look_scale, self.perspective_is_on)

    def __repr__(self):
        theta, phi, alpha = self.get_orientation_angles()
        return f"Camera.from_angles(theta={theta}, phi={phi}, alpha={alpha}, scale={self.look_scale}, look_at={self.look_at.tolist()}, perspective={self.perspective_is_on})"

    def _complete_basis(self, depth) -> tuple[np.ndarray, np.ndarray]:
        z = np.array([0, 0, 1], dtype=np.float32)
        y = np.array([0, 1, 0], dtype=np.float32)
        depth = np.asarray(depth)
        b = z if abs(np.sum(depth * z)) < abs(np.sum(depth * y)) else y
        r = np.cross(depth, b)
        r /= np.sqrt(np.sum(r**2))
        u = np.cross(r, depth)
        u /= np.sqrt(np.sum(u**2))
        return u, r

    def set_orientation_by_angles(self, theta, phi, alpha):
        """Set camera basis in terms of three angles.

        Arguments:
            theta - polar angle of viewing direction,
            phi -  azimuthal angle of viewing direction.
            alpha - angle between direction up and polar axis (z).
        """
        depth = np.array(
            [np.cos(phi) * np.sin(theta), np.sin(phi) * np.sin(theta), np.cos(theta)],
            dtype=np.float32,
        )
        u, r = self._complete_basis(depth)
        up = np.cos(alpha) * u + np.sin(alpha) * r
        return self.set_basis(depth=depth, up=up)

    def get_orientation_angles(self):
        """Inverse function to `Camera.set_orientation_by_angles`."""
        depth = self.look_direction
        up = self.look_up
        theta = np.arccos(depth[2])
        phi = np.arctan2(depth[1], depth[0])
        u, r = self._complete_basis(depth)
        alpha = np.arctan2(np.sum(r * up), np.sum(u * up))
        return theta, phi, alpha

    def snap_orientation(self) -> "Camera":
        self.look_direction = snap_vector(self.look_direction)
        self.look_up = snap_vector(self.look_up)
        self.normalize()
        return self

    def set_predefined_direction(self, preset: int) -> "Camera":
        if preset == 2:
            depth = [0, 0, -1]
            up = [0, 1, 0]
        elif preset == 1:
            depth = [0, 1, 0]
            up = [0, 0, 1]
        elif preset == 0:
            depth = [-1, 0, 0]
            up = [0, 0, -1]
        elif preset == 3:
            depth = [-1, 1, -1]
            up = [0, 0, 1]
        else:
            raise ValueError(f"Preset `{preset}` is unknown")
        self.set_basis(depth=depth, up=up)
        return self

    def reset_FOV(self, system: System) -> "Camera":
        """
        Set default field of view.
        """
        self.reset_center_of_FOV(system)
        self.reset_direction_of_FOV(system)
        return self

    def reset_center_of_FOV(self, system: System) -> "Camera":
        self.look_at = np.dot(np.asarray(system.size) - 1, system.primitives) / 2
        return self

    def reset_direction_of_FOV(self, system: System) -> "Camera":
        primitives = [p for p in system.primitives]
        if len(primitives) < 3:
            primitives.append(np.array([0.0, 0.0, 1.0]))
        self.look_direction = -np.cross(primitives[0], primitives[2])
        self.look_direction /= np.linalg.norm(self.look_direction)
        self.look_up = primitives[2]
        self.look_up /= np.linalg.norm(self.look_up)
        self.look_scale = np.linalg.norm(self.look_at) / np.sqrt(2)
        return self

    def adjust_fov(self, system: System) -> "Camera":
        self.reset_center_of_FOV(system)
        return self

    def adjust_scale(self, system: System, factor: float = 1.1) -> float:
        """
        Adjust scale to completely put the `system` to the screen.
        Return recommended aspect ratio for the image.
        """
        b = system.primitives * system.size[:, None]  # Box sides.
        v = np.stack(
            [
                0 * b[0] + 0 * b[1] + 0 * b[2],
                0 * b[0] + 0 * b[1] + 1 * b[2],
                0 * b[0] + 1 * b[1] + 0 * b[2],
                0 * b[0] + 1 * b[1] + 1 * b[2],
                1 * b[0] + 0 * b[1] + 0 * b[2],
                1 * b[0] + 0 * b[1] + 1 * b[2],
                1 * b[0] + 1 * b[1] + 0 * b[2],
                1 * b[0] + 1 * b[1] + 1 * b[2],
            ],
            axis=0,
        )  # Box vertices
        p = (v[None, :, :] + system.representatives[:, None, :]).reshape(
            (-1, 3)
        )  # Extreme points.
        p = p - self.look_at[None, :]  # Shift origin to the center of view.
        # Project to screen plane.
        x = np.einsum("pk,k->p", v, self.look_right)
        y = np.einsum("pk,k->p", v, self.look_up)
        width = np.max(np.abs(x))
        height = np.max(np.abs(y))
        aspectratio = width / height
        self.look_scale = 0.5 * max(width, height) * factor
        return aspectratio

    @property
    def look_right(self):
        """
        Camera orientation is defined by:
        `self.look_direction` - where we look at (OZ axis),
        `self.look_up` - direction for screen's OY axis,
        `self.look_right` - direction for screen's OX axis.
        """
        horizont = -np.cross(self.look_up, self.look_direction)
        return horizont / np.linalg.norm(horizont)

    @property
    def scale(self):
        return (
            float(self.MAGA) / self.look_scale
            if self.perspective_is_on
            else 1.0 / self.look_scale
        )

    def camera_position(self):
        return self.look_at - self.look_direction / self.scale * 2

    def get_view_matrix(self):
        """
        Return ModelView matrix (should be transposed to use in GL).
        """
        shift = np.eye(4, dtype=np.float32)
        shift[:3, 3] = -self.look_at
        rotate = np.zeros((4, 4), dtype=np.float32)
        rotate[3, 3] = 1
        rotate[0, :3] = self.look_right
        rotate[1, :3] = self.look_up
        rotate[2, :3] = self.look_direction
        scale = np.eye(4, dtype=np.float32)
        scale[0, 0] = scale[1, 1] = scale[2, 2] = self.scale
        modelview = np.dot(scale, np.dot(rotate, shift))
        return modelview

    def get_projection_matrix(self, aspectratio=None):
        """
        Compute ViewProjection matrix (should be transposed to use in GL).
        """
        projection = np.eye(4, dtype=np.float32)
        if aspectratio > 1:
            projection[1, 1] = aspectratio
        else:
            projection[0, 0] = 1.0 / aspectratio
        projection[2, 2] = 1.0 / self.scale / self.depth
        if self.perspective_is_on:
            # w(z)=az+b
            # w(-1)=-a+b=1/mag, w(1)=a+b=mag
            projection[3, 3] = self.MAGA
            projection[3, 2] = self.MAG - self.MAGA
        return projection

    def move_field_of_view(self, x=0, y=0, z=0, rate=0.01) -> "Camera":
        self.look_at += x * self.look_right * self.look_scale * rate
        self.look_at += y * self.look_up * self.look_scale * rate
        self.look_at += z * self.look_direction * self.look_scale * rate
        return self

    def rotate_field_of_view(self, x=0, y=0, rate=0.01 * 2 * np.pi) -> "Camera":
        self.look_direction, self.look_up = (
            self.look_direction - x * rate * self.look_right - y * rate * self.look_up,
            self.look_up + y * rate * self.look_direction,
        )
        return self.normalize()

    def rotate_around(self, x=0, y=0, z=0, rate=0.01 * 2 * np.pi) -> "Camera":
        s, c = np.sin(rate), np.cos(rate)
        self.look_direction, self.look_up = (
            c * self.look_direction + x * s * self.look_up,
            -x * s * self.look_direction + c * self.look_up,
        )
        self.look_direction = c * self.look_direction + y * s * self.look_right
        self.look_up = c * self.look_up + z * s * self.look_right
        return self.normalize()

    def set_basis(self, depth=None, up=None) -> "Camera":
        self.look_direction = np.array(depth, dtype=np.float32)
        self.look_up = np.array(up, dtype=np.float32)
        return self.normalize()

    def normalize(self) -> "Camera":
        self.look_direction /= np.linalg.norm(self.look_direction)
        self.look_up -= np.sum(self.look_direction * self.look_up) * self.look_direction
        self.look_up /= np.linalg.norm(self.look_up)
        return self

    def zoom(self, v, rate=0.01):
        if v > 0:
            self.look_scale /= 1 + rate
        else:
            self.look_scale *= 1 + rate

    def toggle_perspective(self):
        self.perspective_is_on = not self.perspective_is_on


class UnitCamera(Camera):
    def __init__(self):
        self.depth = 10
        self.MAG = 1.0
        self.MAGA = (self.MAG + 1.0 / self.MAG) / 2
        self.perspective_is_on = True
        self.reset_FOV()

    def reset_center_of_FOV(self):
        self.look_at = np.array([0, 0, 0], dtype=np.float32)

    def reset_direction_of_FOV(self):
        self.look_direction = np.array([0, 0, -1], dtype=np.float32)
        self.look_up = np.array([0, 1, 0], dtype=np.float32)
        self.look_scale = 1

    def move_field_of_view(self, x=None, y=None, z=None, rate=0.01):
        return



import moderngl as mgl
import numpy as np

from ..core import System
from .buffer import Field3DBuffer, Field1DBuffer, IntBuffer
from .sphere import SphereDotRender
from .vector import VectorRender
from .charge import ChargeRender
from .flow import FlowRender, AstraRender
from .isolines import IsolinesRender
from .camera import Camera, UnitCamera
from ..core import NumericProperty, BooleanProperty, ListProperty, PropertiesCollection
from .level import LevelRender
from .axes import AxesRender


class Scene:
    """Multiple renders joined into a scene."""

    def __init__(self, mglctx: mgl.Context, system: System, camera: Camera = None):
        """
        Arguments:
            `mglctx` - OpenGL context.
            `system` - lattice description.
            `camera` - initial camera orientation.
        """
        self._system = system
        self._ctx = mglctx
        self._camera = UnitCamera() if camera is None else camera
        assert isinstance(self._system, System)
        assert isinstance(self._ctx, mgl.Context)
        assert isinstance(self._camera, Camera)

    def __getitem__(self, key):
        props_by_render = self.list_properties()
        if key not in props_by_render:
            raise KeyError(
                f"Render `{key}` is unknown. To see list of renders call `yourscene.print_properties()`."
            )
        return PropertiesCollection(props_by_render[key])

    def upload(self, state: np.ndarray = None):
        """Set director field to render."""
        raise NotImplementedError

    def upload_mask(self, mask: np.ndarray = None):
        """Specify grid points to show.

        Renders showing individual vectors/directors at grid points can remove some of the points
        according to the mask.
        """
        pass

    def render(self):
        """Call to create image in earlier created OpenGL context."""
        raise NotImplementedError

    @property
    def camera(self) -> Camera:
        """Common camera for all renders."""
        return self._camera

    def set_camera(self, camera):
        """Change camera object."""
        self._camera = camera

    def list_properties(self):
        """Get all properties available in the scene."""
        return {}

    def list_actions(self):
        """Get all available actions."""
        return {}

    def print_properties(self):
        props_by_render = self.list_properties()
        for render, props in props_by_render.items():
            print(f"'{render:}'")
            for n, prop in enumerate(props):
                print(f"  {n}: {prop}")

    @property
    def detalization(self):
        """Detalization level.

        Higher levels of detalization give smoother surfaces, but render much slower.
        """
        return 1.0

    @detalization.setter
    def detalization(self, value):
        pass


##################################################################################################


class EmptyScene(Scene):
    """Basic scene containing no renders except for background and coordinate axes."""

    def __init__(self, background=(1, 1, 1), *vargs, **kwargs):
        super().__init__(*vargs, **kwargs)
        self._background = background
        self.preferable_size = np.min(np.linalg.norm(self._system.primitives, axis=-1))

        self._axes = AxesRender(mglctx=self._ctx)
        self.show_axes = True

    def upload(self, state=None, **kwargs):
        pass

    def render(self):
        self._ctx.clear(*self._background)

    @property
    def axes_render(self):
        return self._axes

    def render_axes(self, aspectratio):
        if not self.show_axes:
            return
        self._axes.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
        )


#################################################################################


class OrthoScene(EmptyScene):
    """Common ancestor for renders of 3d objects."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._camera.perspective_is_on = False

    def compute_camera_basis(self, axis):
        """Compute natural basis for camera looking along `axis`.

        Similar functionality is given by `Camera.set_predefined_direction`.
        """
        assert self._system.dim == 3
        size = self._system.size
        primitives = self._system.primitives
        if primitives.shape[0] == 3:
            basis = primitives
        elif primitives.shape[0] == 2:
            basis = np.empty((3, 3), dtype=np.float32)
            basis[:2] = primitives
            basis[2] = np.cross(basis[0], basis[1])
        else:
            assert False

        bxis = (axis - 1) % 3
        cxis = (axis - 2) % 3
        if size[bxis] < size[cxis]:
            depth = basis[axis]
            up = -basis[bxis]
        else:
            depth = basis[axis]
            up = basis[cxis]
        return (depth, up)


##########################################################################################################


class SphereDotScene(EmptyScene):
    """Render all directors on Block sphere."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._buffer = Field3DBuffer(self._ctx, self._system)
        self._render = SphereDotRender(
            mglctx=self._ctx, system=self._system, stateBuffer=self._buffer
        )

    def upload(self, state: np.ndarray = None, **kwargs):
        self._buffer.upload(state)

    def render(self, aspectratio=None):
        self._ctx.clear(*self._background)
        self._ctx.enable(mgl.DEPTH_TEST)
        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
        )

    def list_properties(self):
        return {"Ball": self._render.list_properties()}


##########################################################################################################


class VectorScene(EmptyScene):
    """Render director field as vectors/cylinders at every grid point."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._state = None
        self._m_state = self._system.field3D()
        self._buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._coordinates_buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._mask_buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._scalar_mask_buffer = Field1DBuffer(mglctx=self._ctx, system=self._system)

        self._render = VectorRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            coordinatesBuffer=self._coordinates_buffer,
            maskBuffer=self._mask_buffer,
            scalarMaskBuffer=self._scalar_mask_buffer,
        )
        self._render.base_size = self.preferable_size

    @property
    def vector_render(self):
        return self._render

    def get_use_energy_mask(self):
        return self._render.use_scalar_mask

    def set_use_energy_mask(self, value):
        self._render.use_scalar_mask = value
        self.update_scalar_mask()

    use_energy_mask = property(get_use_energy_mask, set_use_energy_mask)

    def update_scalar_mask(self):
        # print("update_scalar_mask")
        if not self._render.use_scalar_mask:
            return
        if self._state is None:
            return
        self._render.use_mask = True

        m_scalar1 = self._system.field1D()
        m_scalar2 = self._system.field1D()
        self._m_state.upload(self._state).energy_contributions(
            dm=m_scalar1, heisenberg=m_scalar2
        )
        scalar = -m_scalar1.download() - m_scalar2.download()
        mn, mx = np.min(scalar), np.max(scalar)
        scalar = (scalar - mn) / (mx - mn)
        scalar = np.power(scalar, 32)
        # scalar = (np.exp(scalar)-1)/np.exp(1)

        # scalar = -self._m_state.upload(self._state).energy().download()
        # scalar = np.abs(self._m_state.upload(self._state).top_charge().download())
        self._scalar_mask_buffer.upload(scalar)
        self._render.min_scalar = 0  # np.min(scalar)
        self._render.max_scalar = 1  # np.max(scalar)
        # print(self._render.min_scalar, self._render.max_scalar, self._render.use_scalar_mask)

    def upload(self, state: np.ndarray = None):
        self._state = state
        if not state is None:
            self._buffer.upload(state)
            self.update_scalar_mask()

    def upload_mask(self, mask: np.ndarray = None):
        self._render.use_scalar_mask = False
        if mask is None:
            self._render.use_mask = False
        else:
            self._mask_buffer.upload(mask)
            self._render.use_mask = True

    def render(self, aspectratio=None):
        self._ctx.clear(*self._background)
        self._ctx.enable_only(mgl.DEPTH_TEST)
        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    def list_properties(self):
        return {
            "Cones": self._render.list_properties(),
            "Mask": [
                BooleanProperty(
                    self, "use_energy_mask", title="Energy mask", reset=True
                ),
            ],
        }

    @property
    def detalization(self):
        return self._render.detalization

    @detalization.setter
    def detalization(self, value):
        self._render.detalization = value


##########################################################################################################


class LayerScene(OrthoScene):
    """Show sections of director field by a coordinate axis.

    Very similar to `VectorScene` but show only gird points sharing one coordinate defined by `axis` property.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._coordinates_buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._mask_buffer = self._coordinates_buffer  # Not used
        self._scalar_mask_buffer = self._coordinates_buffer

        self._render = VectorRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            coordinatesBuffer=self._coordinates_buffer,
            maskBuffer=self._mask_buffer,
            scalarMaskBuffer=self._scalar_mask_buffer,
        )
        self._render.min_size = 0
        self._render.base_size = self.preferable_size
        self._render.min_threshold = 0
        self._render.power_size = 1

        self.axis = 2
        self.show_axes = True

    @property
    def vector_render(self):
        return self._render

    def get_axis(self):
        return self._axis

    def set_axis(self, value):
        self._axis = int(value)
        self.layers_count = self._render.size[self._axis]
        self.layer = self.layers_count / 2

        basis = self.compute_camera_basis(self._axis)
        # print("Camera orientation:", basis)
        self._camera.set_basis(*basis)

    def get_layer(self):
        return self._layer

    def set_layer(self, value):
        self._layer = int(value)
        mn = 0 * np.array(self._render.size)
        mx = np.array(self._render.size)
        mn[self._axis] = self._layer
        mx[self._axis] = self._layer
        self._render.clip_min = mn
        self._render.clip_max = mx

    axis = property(get_axis, set_axis)
    layer = property(get_layer, set_layer)

    def upload(self, state: np.ndarray = None):
        if not state is None:
            self._buffer.upload(state)

    def upload_mask(self, mask: np.ndarray = None):
        pass

    def render(self, aspectratio: float = None):
        self._ctx.clear(*self._background)
        self._ctx.enable_only(mgl.DEPTH_TEST)
        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    def list_properties(self):
        return {
            "Filter": [
                NumericProperty(
                    self, "axis", title="Axis", min=0, max=2, count=3, reset=True
                ),
                NumericProperty(
                    self,
                    "layer",
                    title="Layer",
                    min=0,
                    max=self.layers_count - 1,
                    count=self.layers_count,
                ),
            ],
            # "Cones": self._render.list_properties(),
            "Cones": [
                ListProperty(
                    self._render,
                    "mesh_shape",
                    title="Shape",
                    values=sorted(self._render.MESH_NAMES.keys()),
                ),
                # BooleanProperty(self._render, "director_mode", title="Director mode"),
                BooleanProperty(self._render, "negate", title="Negate spin"),
                NumericProperty(
                    self._render, "width", title="Width", min=0.01, max=2.0, count=100
                ),
                NumericProperty(
                    self._render,
                    "color_shift",
                    title="Hue",
                    min=0.0,
                    max=1.0,
                    count=360,
                ),
                NumericProperty(
                    self._render,
                    "saturation_mag",
                    title="Saturation",
                    min=1.0,
                    max=10.0,
                    count=500,
                ),
                # NumericProperty(self, "min_size", title="Min. size", min=0.0, max=2.0, count=30),
                NumericProperty(
                    self._render,
                    "max_size",
                    title="Max. size",
                    min=0.0,
                    max=2.0,
                    count=30,
                ),
                # NumericProperty(self._render, "power_size", title="Steepness", min=0.0, max=2.0, count=30),
                # NumericProperty(self._render, "min_threshold", title="Min. threshold", min=0.0, max=1.0, count=500),
                # NumericProperty(self._render, "detalization", title="Detalization", min=1., max=10., count=10),
            ],
        }

    @property
    def detalization(self):
        return self._render.detalization

    @detalization.setter
    def detalization(self, value):
        self._render.detalization = value


##########################################################################################################


class LayerAndSurfaceScene(LayerScene):
    """Mix of LayerScene and SurfaceScene."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        idx = self._system.get_inner_simplices()
        self._idx = IntBuffer(mglctx=self._ctx, count=4 * idx.shape[0])
        self._idx.upload(idx.flatten())

        self._level_render = FlowRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            indexBuffer=self._idx,
            coordinatesBuffer=self._coordinates_buffer,
        )

        self._camera.perspective_is_on = True

    @property
    def vector_render(self):
        return self._render

    @property
    def isosurface_render(self):
        return self._level_render

    def render(self, aspectratio: float = None):
        self._ctx.clear(
            *(
                [1.0, 1.0, 1.0, 1.0]
                if self._level_render.inverted
                else [0.0, 0.0, 0.0, 1.0]
            )
        )
        self._ctx.enable_only(mgl.DEPTH_TEST)

        self._render.color_shift = self._level_render.color_shift
        self._render.lighting = self._level_render.lighting
        self._render.negate = self._level_render.negate

        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self._level_render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    def list_properties(self):
        return {
            "Layer": [
                NumericProperty(
                    self, "axis", title="Axis", min=0, max=2, count=3, reset=True
                ),
                NumericProperty(
                    self,
                    "layer",
                    title="Layer",
                    min=0,
                    max=self.layers_count - 1,
                    count=self.layers_count,
                ),
                ListProperty(
                    self._render,
                    "mesh_shape",
                    title="Shape",
                    values=sorted(self._render.MESH_NAMES.keys()),
                ),
                NumericProperty(
                    self._render, "width", title="Width", min=0.01, max=2.0, count=100
                ),
                NumericProperty(
                    self._render,
                    "max_size",
                    title="Max. size",
                    min=0.0,
                    max=2.0,
                    count=30,
                ),
            ],
            "Isosurface": [
                BooleanProperty(
                    self._level_render, "neglect_z", title="Ignore z-projection"
                ),
                ListProperty(
                    self._level_render,
                    "angle_name",
                    title="Constant",
                    values=self._level_render.ANGLENAMES,
                ),
                NumericProperty(
                    self._level_render,
                    "level",
                    title="Level",
                    min=-1.0,
                    max=1.0,
                    count=1000,
                ),
            ],
            "Shared": [
                BooleanProperty(self._level_render, "inverted", title="Invert colors"),
                BooleanProperty(self._level_render, "lighting", title="Lighting"),
                BooleanProperty(self._level_render, "negate", title="Negate"),
                NumericProperty(
                    self._level_render,
                    "color_shift",
                    title="Hue",
                    min=0.0,
                    max=1.0,
                    count=360,
                ),
            ],
        }


##########################################################################################################


class LayerAndIsolinesScene(LayerScene):
    """Mix of LayerScene and IsolinesScene."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        idx = self._system.get_inner_simplices()
        self._idx = IntBuffer(mglctx=self._ctx, count=4 * idx.shape[0])
        self._idx.upload(idx.flatten())

        self._level_render = IsolinesRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            indexBuffer=self._idx,
            coordinatesBuffer=self._coordinates_buffer,
        )

        self._camera.perspective_is_on = True

    @property
    def vector_render(self):
        return self._render

    @property
    def isolines_render(self):
        return self._level_render

    def render(self, aspectratio: float = None):
        self._ctx.clear(
            *(
                [1.0, 1.0, 1.0, 1.0]
                if self._level_render.inverted
                else [0.0, 0.0, 0.0, 1.0]
            )
        )
        self._ctx.enable_only(mgl.DEPTH_TEST)

        self._render.color_shift = self._level_render.color_shift
        self._render.lighting = self._level_render.lighting
        self._render.negate = self._level_render.negate

        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self._ctx.line_width = 3
        self._level_render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    def list_properties(self):
        return {
            "Layer": [
                NumericProperty(
                    self, "axis", title="Axis", min=0, max=2, count=3, reset=True
                ),
                NumericProperty(
                    self,
                    "layer",
                    title="Layer",
                    min=0,
                    max=self.layers_count - 1,
                    count=self.layers_count,
                ),
                ListProperty(
                    self._render,
                    "mesh_shape",
                    title="Shape",
                    values=sorted(self._render.MESH_NAMES.keys()),
                ),
                NumericProperty(
                    self._render, "width", title="Width", min=0.01, max=2.0, count=100
                ),
                NumericProperty(
                    self._render,
                    "max_size",
                    title="Max. size",
                    min=0.0,
                    max=2.0,
                    count=30,
                ),
            ],
            "Isolines": [
                BooleanProperty(
                    self._level_render, "neglect_z", title="Ignore z-projection"
                ),
                NumericProperty(
                    self._level_render,
                    "phi_shift",
                    title="Polar angle offset",
                    min=-1.0,
                    max=1.0,
                    count=1000,
                ),
                NumericProperty(
                    self._level_render,
                    "theta_shift",
                    title="Azimuthal angle offset",
                    min=-1.0,
                    max=1.0,
                    count=1000,
                ),
                NumericProperty(
                    self._level_render,
                    "npetals",
                    title="Polar angle substeps",
                    min=1,
                    max=10,
                    count=10,
                ),
                NumericProperty(
                    self._level_render,
                    "ntheta",
                    title="Azimuthal angle substeps",
                    min=1,
                    max=5,
                    count=5,
                ),
                NumericProperty(
                    self._level_render,
                    "width",
                    title="Width",
                    min=0.0,
                    max=0.02,
                    count=1000,
                ),
            ],
            "Shared": [
                BooleanProperty(self._level_render, "inverted", title="Invert colors"),
                BooleanProperty(self._level_render, "lighting", title="Lighting"),
                BooleanProperty(self._level_render, "negate", title="Negate"),
                NumericProperty(
                    self._level_render,
                    "color_shift",
                    title="Hue",
                    min=0.0,
                    max=1.0,
                    count=360,
                ),
            ],
        }


##########################################################################################################


class CubeScene(OrthoScene):
    """Show director field as vectors similar to VectorScene with cubic cut."""

    def __init__(self, *vargs, **kwargs):
        super().__init__(*vargs, **kwargs)
        self._buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._coordinates_buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._mask_buffer = self._coordinates_buffer  # Not used
        self._scalar_mask_buffer = self._coordinates_buffer

        self._render = VectorRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            coordinatesBuffer=self._coordinates_buffer,
            maskBuffer=self._mask_buffer,
            scalarMaskBuffer=self._scalar_mask_buffer,
        )
        self._render.min_size = 0
        self._render.base_size = self.preferable_size
        self._render.min_threshold = 0
        self._render.power_size = 1
        self._render.clip_invert = True

        self._section = np.array(self._render.size, dtype=np.int32) / 2
        self._reverse = np.array(
            [
                False,
            ]
            * 3
        )
        self.update_parameters()
        self._camera.set_basis(*self.compute_camera_basis())

    def update_parameters(self):
        self._render.clip_min = [
            0 if self._reverse[n] else self._section[n] for n in range(3)
        ]
        self._render.clip_max = [
            self._section[n] + 1 if self._reverse[n] else self._render.size[n]
            for n in range(3)
        ]

    def get_section(self, n):
        return self._section[n]

    def set_section(self, n, value):
        self._section[n] = int(value)
        self.update_parameters()

    def get_reverse(self, n):
        return self._reverse[n]

    def set_reverse(self, n, value):
        self._reverse[n] = bool(value)
        self.update_parameters()
        self._camera.set_basis(*self.compute_camera_basis())

    section_x = property(
        lambda self: self.get_section(0), lambda self, v: self.set_section(0, v)
    )
    section_y = property(
        lambda self: self.get_section(1), lambda self, v: self.set_section(1, v)
    )
    section_z = property(
        lambda self: self.get_section(2), lambda self, v: self.set_section(2, v)
    )

    reverse_x = property(
        lambda self: self.get_reverse(0), lambda self, v: self.set_reverse(0, v)
    )
    reverse_y = property(
        lambda self: self.get_reverse(1), lambda self, v: self.set_reverse(1, v)
    )
    reverse_z = property(
        lambda self: self.get_reverse(2), lambda self, v: self.set_reverse(2, v)
    )

    def compute_camera_basis(self):
        d = np.array([1 if self._reverse[n] else -1 for n in range(3)])
        primitives = self._system.primitives

        if primitives.shape[0] == 3:
            basis = primitives
        elif primitives.shape[0] == 2:
            basis = np.empty((3, 3), dtype=np.float32)
            basis[:2] = primitives
            basis[2] = np.cross(basis[0], basis[1])
        else:
            assert False

        depth = np.sum(basis * d[:, None], axis=-1)
        d[1] *= -1
        up = np.sum(basis * d[:, None], axis=-1)
        return (depth, -up)

    def upload(self, state: np.ndarray = None):
        if not state is None:
            self._buffer.upload(state)

    def upload_mask(self, mask: np.ndarray = None):
        pass

    def render(self, aspectratio=None):
        self._ctx.clear(*self._background)
        self._ctx.enable_only(mgl.DEPTH_TEST)
        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    def list_properties(self):
        return {
            "Filter": [
                NumericProperty(
                    self,
                    "section_x",
                    title="X-section",
                    min=0,
                    max=self._render.size[0],
                    count=self._render.size[0] + 1,
                ),
                NumericProperty(
                    self,
                    "section_y",
                    title="Y-section",
                    min=0,
                    max=self._render.size[1],
                    count=self._render.size[1] + 1,
                ),
                NumericProperty(
                    self,
                    "section_z",
                    title="Z-section",
                    min=0,
                    max=self._render.size[2],
                    count=self._render.size[2] + 1,
                ),
                BooleanProperty(self, "reverse_x", title="Reverse X"),
                BooleanProperty(self, "reverse_y", title="Reverse Y"),
                BooleanProperty(self, "reverse_z", title="Reverse Z"),
            ],
            "Cones": [
                NumericProperty(
                    self._render, "max_size", title="Size", min=0.0, max=2.0, count=30
                ),
            ],
        }

    @property
    def detalization(self):
        return self._render.detalization

    @detalization.setter
    def detalization(self, value):
        self._render.detalization = value


##########################################################################################################


class ChargeScene(OrthoScene):
    """Show distribution of topological charge.

    Charge computed on each coordinate plane in shown by RGB components of color.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._coordinates_buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)

        idx = self._system.get_inner_triangles(
            triangulation=self._system.default_trivial_cell_triangulation()
        )
        self._idx = IntBuffer(mglctx=self._ctx, count=3 * idx.shape[0])
        self._idx.upload(idx.flatten())

        self._render = ChargeRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            indexBuffer=self._idx,
            coordinatesBuffer=self._coordinates_buffer,
        )

        self.update_parameters()

    def update_parameters(self):
        pass

    def upload(self, state: np.ndarray = None):
        if not state is None:
            self._buffer.upload(state)

    def render(self, aspectratio: float = None):
        self._ctx.clear(
            *([1.0, 1.0, 1.0, 1.0] if self._render.inverted else [0.0, 0.0, 0.0, 1.0])
        )
        self._ctx.enable_only(mgl.DEPTH_TEST)
        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    def list_properties(self):
        return {"Charge": self._render.list_properties()}


##########################################################################################################


class FlowScene(EmptyScene):
    """Show isosurface of director field."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self._system.dim != 3:
            raise NotImplementedError("Only 3D systems are supported.")

        self._buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._coordinates_buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)

        idx = self._system.get_inner_simplices()
        self._idx = IntBuffer(mglctx=self._ctx, count=4 * idx.shape[0])
        self._idx.upload(idx.flatten())

        self._render = FlowRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            indexBuffer=self._idx,
            coordinatesBuffer=self._coordinates_buffer,
        )

        self._camera.perspective_is_on = True

        self.update_parameters()

    def update_parameters(self):
        pass

    @property
    def surface_render(self):
        return self._render

    def upload(self, state: np.ndarray = None):
        if not state is None:
            self._buffer.upload(state)

    def render(self, aspectratio: float = None):
        self._ctx.clear(
            *([1.0, 1.0, 1.0, 1.0] if self._render.inverted else [0.0, 0.0, 0.0, 1.0])
        )
        self._ctx.enable_only(mgl.DEPTH_TEST)
        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    def list_properties(self):
        return {"Level surface": self._render.list_properties()}


##########################################################################################################


class LevelScene(EmptyScene):
    """Show isusurface of the director field."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self._system.dim != 3:
            raise NotImplementedError("Only 3D systems are supported.")

        self._buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._coordinates_buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)

        idx = self._system.get_inner_simplices()
        self._idx = IntBuffer(mglctx=self._ctx, count=4 * idx.shape[0])
        self._idx.upload(idx.flatten())

        self._render = LevelRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            indexBuffer=self._idx,
            coordinatesBuffer=self._coordinates_buffer,
        )

        self._camera.perspective_is_on = True

        self.update_parameters()

    def update_parameters(self):
        pass

    def upload(self, state: np.ndarray = None):
        if not state is None:
            self._buffer.upload(state)

    def render(self, aspectratio: float = None):
        self._ctx.clear(
            *([1.0, 1.0, 1.0, 1.0] if self._render.inverted else [0.0, 0.0, 0.0, 1.0])
        )
        self._ctx.enable_only(mgl.DEPTH_TEST)
        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    def list_properties(self):
        return {"Level surface": self._render.list_properties()}


##########################################################################################################


class AstraScene(EmptyScene):
    """Show multiple isusurfaces of the director field."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self._system.dim != 3:
            raise NotImplementedError("Only 3D systems are supported.")

        self._buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._coordinates_buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)

        idx = self._system.get_inner_simplices()
        self._idx = IntBuffer(mglctx=self._ctx, count=4 * idx.shape[0])
        self._idx.upload(idx.flatten())

        self._render = AstraRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            indexBuffer=self._idx,
            coordinatesBuffer=self._coordinates_buffer,
        )

        self._camera.perspective_is_on = True

        self.update_parameters()

    def update_parameters(self):
        pass

    def upload(self, state: np.ndarray = None):
        if not state is None:
            self._buffer.upload(state)

    def render(self, aspectratio: float = None):
        self._ctx.clear(
            *([1.0, 1.0, 1.0, 1.0] if self._render.inverted else [0.0, 0.0, 0.0, 1.0])
        )
        self._ctx.enable_only(mgl.DEPTH_TEST)
        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    @property
    def surface_render(self):
        return self._render

    def list_properties(self):
        return {"Astra": self._render.list_properties()}


##########################################################################################################


class IsolinesScene(EmptyScene):
    """Show isolines of the director field."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self._system.dim != 3:
            raise NotImplementedError("Only 3D systems are supported.")

        self._buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)
        self._coordinates_buffer = Field3DBuffer(mglctx=self._ctx, system=self._system)

        idx = self._system.get_inner_simplices()
        self._idx = IntBuffer(mglctx=self._ctx, count=4 * idx.shape[0])
        self._idx.upload(idx.flatten())

        self._render = IsolinesRender(
            mglctx=self._ctx,
            system=self._system,
            stateBuffer=self._buffer,
            indexBuffer=self._idx,
            coordinatesBuffer=self._coordinates_buffer,
        )

        self._camera.perspective_is_on = True

        self.update_parameters()

    @property
    def isolines_render(self):
        return self._render

    def update_parameters(self):
        pass

    def upload(self, state: np.ndarray = None):
        if not state is None:
            self._buffer.upload(state)

    def render(self, aspectratio: float = None):
        self._ctx.clear(
            *([1.0, 1.0, 1.0, 1.0] if self._render.inverted else [0.0, 0.0, 0.0, 1.0])
        )
        self._ctx.enable_only(mgl.DEPTH_TEST)
        self._ctx.line_width = 3
        self._render.render(
            modelview=self.camera.get_view_matrix(),
            projection=self.camera.get_projection_matrix(aspectratio=aspectratio),
            camera_position=self.camera.camera_position(),
        )

        self.render_axes(aspectratio=aspectratio)

    @property
    def surface_render(self):
        return self._render

    def list_properties(self):
        return {"Isolines": self._render.list_properties()}


##########################################################################################################

# List of all available scenes.
SCENES = {
    "Vectors": VectorScene,
    "Level surface": FlowScene,
    # "Level surface-2": LevelScene,
    "Astra": AstraScene,
    "Isolines": IsolinesScene,
    "Top. charge": ChargeScene,
    "Layer": LayerScene,
    "Cube": CubeScene,
    "Ball": SphereDotScene,
    "Layer and isosurface": LayerAndSurfaceScene,
    "Layer and isolines": LayerAndIsolinesScene,
}


def print_scenes():
    print("Scenes:", ", ".join(SCENES))



import moderngl as mgl
import numpy as np

from .render import *
from .buffer import Float3DBuffer, IntBuffer
from ..core.property import NumericProperty, ListProperty, BooleanProperty


######################################################################################################

# GLSL functions for lattice points computation. For future usage.
#
# uniform ivec4 u_size; // Number of spins per each asis+number of atoms per unit cell.
# uniform mat3 u_generators; // Lattice generators.
# uniform vec3 u_representatives[]; // Atoms in the the unit cell.
#
# ivec4 reshape(int id) {
#     ivec4 stride;
#     stride.t = 1;
#     stride.z = stride.t*u_size.t;
#     stride.y = stride.z*u_size.z;
#     stride.x = stride.t*u_size.y;
#     return (id / stride) % u_size;
# }
#
# pos get_coordinates() {
#     ivec4 idx = reshape(gl_VertexID);
#     return u_generators*idx.xyz + u_representatives[idx.t];
# }


class ChargeRender(StateRender):
    VERTEX = """
            #version 330

            const float M_PI = 3.1415926535897932384626433832795;

            uniform mat4 u_view;
            uniform mat4 u_projection;
            uniform vec3 u_camera; // camera position in world coordinates

            in vec3 in_center; // center of atom
            in vec3 in_spin; // spin direction

            out VS_OUT {
                vec3 origin; // atom position in world coordiantes
                vec3 spin; // spin direction in spin space
                vec3 light_direction; 
            } vs_out;

            void main() {
                vs_out.origin = in_center;
                vs_out.spin = in_spin;
                vs_out.light_direction = - normalize(vs_out.origin-u_camera);

                gl_Position = u_projection * u_view * vec4(vs_out.origin, 1.0);
            } 
        """
    GEOMETRY = """
            #version 330 core

            layout (triangles) in;
            layout (triangle_strip, max_vertices = 4) out;

            uniform float u_brightness;
            uniform bool u_inverted;

            in VS_OUT {
                vec3 origin; 
                vec3 spin; 
                vec3 light_direction; 
            } vs_in[];

            out vec3 v_normal;
            out vec3 v_position;
            out vec3 v_color;
            out vec3 v_light_direction;

            void main() {
                vec3 ab = vs_in[1].origin - vs_in[0].origin;
                vec3 ac = vs_in[2].origin - vs_in[0].origin;
                v_normal = normalize(cross(ab, ac));

                float top_charge = dot(vs_in[0].spin, cross(vs_in[1].spin, vs_in[2].spin));

                vec3 top_color;

                // 2D charge
                /*
                float plus_intens = clamp(u_brightness*top_charge, 0, 1);
                float minus_intens = clamp(-u_brightness*top_charge, 0, 1);

                if (u_inverted) {
                    top_color = vec3(1-minus_intens, 1-plus_intens-minus_intens, 1-plus_intens);
                } else {
                    top_color = vec3(plus_intens, 0, minus_intens);
                };
                */

                // 3D charge
                vec3 clr = v_normal.yzx*top_charge;
                if (u_inverted) {
                    top_color = 1.0-u_brightness*(clamp(abs(clr),0.,1.)+clamp(-clr.yzx,0.,1.)+clamp(clr.zxy,0.,1.));
                } else {
                    top_color = u_brightness*(clamp(clr.yzx,0.,1.)+clamp(-clr.zxy,0.,1.));
                };

                for (int k=0; k<3; k++) {
                    gl_Position = gl_in[k].gl_Position;
                    v_position = vs_in[k].origin;
                    v_color = top_color;
                    EmitVertex();
                };
                EndPrimitive();
            }   
        """
    FRAGMENT = """
            #version 330

            uniform vec3 v_camera;

            in vec3 v_normal;
            in vec3 v_position;
            in vec3 v_color;
            in vec3 v_light_direction;

            out vec4 f_color;

            void main() {
                f_color = vec4(v_color, 1.0);
            }
        """

    def __init__(
        self,
        coordinatesBuffer: Float3DBuffer = None,
        maskBuffer=None,
        indexBuffer: IntBuffer = None,
        scalarMaskBuffer=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.u_view = self._prog["u_view"]
        self.u_projection = self._prog["u_projection"]

        self._coordinatesBuffer = coordinatesBuffer
        self._coordinatesBuffer.upload(self._system.spin_positions())

        self._idx = indexBuffer

        self.init_vao()

        self.brightness = 3.0
        self.inverted = True

    def init_vao(self):
        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self._state.vbo, "3f/v", "in_spin"),
                (self._coordinatesBuffer.vbo, "3f/v", "in_center"),
            ],
            self._idx.ibo,
        )

    def render(
        self, scales=None, modelview=None, projection=None, camera_position=None
    ):
        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())

        self.u_brightness.value = self.brightness
        self.u_inverted.value = self.inverted

        self._ctx.enable_only(mgl.BLEND)
        self._ctx.blend_func = mgl.ONE, mgl.ONE
        self._ctx.blend_equation = mgl.MIN if self.inverted else mgl.MAX

        self._vao.render(mode=mgl.TRIANGLES, vertices=-1, first=0)

    def list_properties(self):
        return [
            BooleanProperty(self, "inverted", title="Invert colors"),
            NumericProperty(
                self, "brightness", title="Brightness", min=0.0, max=10.0, count=1000
            ),
        ]



from .optics import Function, FunctionTable, ConstantFunction, GaussianFunction, SimplifiedRGB, BlackbodyRadiation, render_cp_multi

__all__ = [
    "Function", "FunctionTable", "ConstantFunction", "GaussianFunction", "SimplifiedRGB", "BlackbodyRadiation", "render_cp_multi",
]


import numpy as np

from ..core import System, log
from .image import ImageContainer
from .camera import Camera
from .scene import LayerScene, FlowScene, IsolinesScene, SCENES


def render_prepare(
    scenecls: str,
    system: System,
    camera_preset: int = 2,
    scalex: bool = False,
    imgsize: int = 1024,
):
    """Prepare scene to offscreen render. single layer render.

    Arguments:
        scenecls: scene class to instantiate. Can be defined by name according to `mgl.print_scenes()`.

    Example:

    ic = render_layer(system) # Create render.
    ic.upload(director) # Specify director field to render.
    ic.save() # Render.
    """

    # Setup camera.
    camera = Camera()
    # Preset 0-2 define axis to look along.
    aspectratio = (
        camera.set_predefined_direction(preset=camera_preset)
        .adjust_fov(system)
        .adjust_scale(system)
    )

    # Define image size
    size = (
        (int(imgsize * aspectratio), imgsize)
        if scalex
        else (imgsize, int(imgsize / aspectratio))
    )

    # Select scene to render.
    if isinstance(scenecls, str):
        scenecls = SCENES[scenecls]

    # Initialize image to render to.
    ic = ImageContainer(
        size=size,
        scene=lambda mglctx: scenecls(mglctx=mglctx, system=system, camera=camera),
    )

    return ic


#############################################################################################################


def render_layer(
    system: System,
    scalex: bool = False,
    imgsize: int = 1024,
    show_axes: bool = False,
    layer: int = 0,
    axis: int = 2,
    mesh: str = "Cylinder",
):
    """Prepare single layer render.

    Example:

    ic = render_layer(system) # Create render.
    ic.upload(director) # Specify director field to render.
    ic.save() # Render.
    """

    # Check parameters.
    if not 0 <= axis < 3:
        raise ValueError(f"Axis should be 0 (x-axis), 1 (y-axis) or 2 (z-axis).")

    if not 0 <= layer < system.size[axis]:
        raise ValueError(
            f"Layer {layer} is outside of allowed range 0 .. {system.size[axis]}"
        )

    # Setup camera.
    camera = Camera()
    # Preset 0-2 define axis to look along.
    aspectratio = (
        camera.set_predefined_direction(preset=axis)
        .adjust_fov(system)
        .adjust_scale(system)
    )
    # log.debug(f"Aspect ratio {aspectratio:.2f}")

    # Define image size
    size = (
        (int(imgsize * aspectratio), imgsize)
        if scalex
        else (imgsize, int(imgsize / aspectratio))
    )

    # Select scene to render.
    scenecls = LayerScene  # Render single layer.

    # Initialize image to render to.
    ic = ImageContainer(
        size=size,
        scene=lambda mglctx: scenecls(mglctx=mglctx, system=system, camera=camera),
    )

    # Tune system parameters if needed.
    ic.scene.axis = axis
    ic.scene.layer = layer

    ic.scene.vector_render.mesh_shape = mesh
    # ic.scene.vector_render.base_size = thinning*0.9
    # ic.scene.vector_render.saturation_mag = mag

    return ic


################################################################################################################################


def render_isosurface(
    system: System,
    scalex: bool = False,
    imgsize: int = 1024,
    show_axes: bool = False,
    preset: int = 2,
    levels: list[float] = [0],
    lighting: bool = True,
    neglect_z: bool = True,
):
    """Prepare iso-surface render.

    Example:

    ic = render_layer(system) # Create render.
    ic.upload(director) # Specify director field to render.
    ic.save() # Render.
    """

    # Setup camera.
    camera = Camera()
    # Preset 0-2 define axis to look along.
    aspectratio = (
        camera.set_predefined_direction(preset=preset)
        .adjust_fov(system)
        .adjust_scale(system)
    )

    # Define image size
    size = (
        (int(imgsize * aspectratio), imgsize)
        if scalex
        else (imgsize, int(imgsize / aspectratio))
    )

    # Select scene to render.
    scenecls = FlowScene  # Render single layer.

    # Initialize image to render to.
    ic = ImageContainer(
        size=size,
        scene=lambda mglctx: scenecls(mglctx=mglctx, system=system, camera=camera),
    )

    ic.scene.surface_render.neglect_z = False
    ic.scene.surface_render.levels = levels
    # ic.scene.surface_render.cutx = lx
    # ic.scene.surface_render.cuty = ly
    # ic.scene.surface_render.cutphi = np.pi/2
    ic.scene.surface_render.lighting = lighting
    ic.scene.surface_render.neglect_z = neglect_z

    return ic


################################################################################################################################


def render_isolines(
    system: System,
    scalex: bool = False,
    imgsize: int = 1024,
    show_axes: bool = False,
    preset: int = 2,
    levels: list[float] = [0],
    lighting: bool = True,
    neglect_z: bool = True,
    thickness: float = 0.002,
    nlines: int = 5,
    z0: float = 0.0,
    phi0: float = 0.0,
):
    """Prepare iso-lines render.

    Arguments:
        thickness # Isoline thickness
        nlines # Number of isolines
        phi0   # Orientation of xy-projection of projector for isolines:
        z0     # z-projection of director for isolines.
               # S=(sqrt(1-z0^2) cos phi0, sqrt(1-z0^2) sin phi0, z0)

    Example:

    ic = render_layer(system) # Create render.
    ic.upload(director) # Specify director field to render.
    ic.save() # Render.
    """

    # Setup camera.
    camera = Camera()
    # Preset 0-2 define axis to look along.
    aspectratio = (
        camera.set_predefined_direction(preset=preset)
        .adjust_fov(system)
        .adjust_scale(system)
    )

    # Define image size
    size = (
        (int(imgsize * aspectratio), imgsize)
        if scalex
        else (imgsize, int(imgsize / aspectratio))
    )

    # Select scene to render.
    scenecls = IsolinesScene  # Render single layer.

    # Initialize image to render to.
    ic = ImageContainer(
        size=size,
        scene=lambda mglctx: scenecls(mglctx=mglctx, system=system, camera=camera),
    )

    ic.scene.isolines_render.neglect_z = False
    ic.scene.isolines_render.levels = levels
    ic.scene.isolines_render.lighting = lighting
    ic.scene.isolines_render.neglect_z = neglect_z
    ic.scene.isolines_render.width = thickness
    # k = np.round( (nlines-1) )#/np.sqrt(1-z0**2) )
    ic.scene.isolines_render.npetals = nlines - 1
    ic.scene.isolines_render.phi_shift = phi0 / (2 * np.pi) * nlines
    ic.scene.isolines_render.theta_shift = np.arcsin(z0)

    ic.scene.show_axes = show_axes
    ic.scene.axes_render.location = (-0.9, -0.6)

    return ic



import logging

log = logging.getLogger("topovec")


def configLog(level=logging.DEBUG):
    from rich.logging import RichHandler

    global log
    log.setLevel(level)
    log.addHandler(RichHandler(rich_tracebacks=True))



import moderngl as mgl
import numpy as np

from .render import Render
from .buffer import Float3DBuffer
from .solid import ArrowMesh

#######################################################################################################


class AxesRender(Render):
    VERTEX = """
                #version 330

                const float M_PI = 3.1415926535897932384626433832795;

                uniform mat4 u_view;
                uniform mat4 u_projection;

                uniform float u_ratio;
                uniform float u_length;
                uniform vec2 u_location;

                in vec3 in_origin; // origin of axis
                in vec3 in_dir; // direction of axis

                in vec3 in_pos; // mesh point in mesh coordinates
                in vec3 in_normal; // normal to mesh at in_pos in mesh coordinates
                in vec3 in_color; // color

                out vec3 v_pos; // vertex position in world coordiantes
                out vec3 v_normal; // normal to surface in world coordinates
                out vec4 v_color;
                out vec3 v_camera;
                out vec3 v_light_direction;

                mat3 rotmat(vec3 z) {
                    vec3 a = cross(z, vec3(1.0,0.0,0.0));
                    vec3 b = cross(z, vec3(0.0,1.0,0.0));
                    vec3 c = (length(a)>length(b)) ? a : b;
                    vec3 x = normalize(c);
                    vec3 y = cross(x, z);
                    return mat3(x, y, z);
                }

                void main() {
                    vec3 dir = normalize(in_dir);
                    mat3 mesh = rotmat(dir);
                    vec3 pos = in_pos*vec3(u_ratio,u_ratio,1);
                    v_pos = in_origin + mesh * pos;

                    mat4 view = mat4(u_view);
                    mat4 projection = mat4(u_projection);
                    view[0] = normalize(view[0]);
                    view[1] = normalize(view[1]);
                    view[2] = normalize(view[2]);
                    view[3] = vec4(0,0,0,1);
                    projection[2][2] = 1;

                    vec4 glpos = projection * view * vec4(v_pos*u_length, 1.0);
                    gl_Position = glpos/glpos.w + vec4(u_location,0,0);

                    v_normal = mesh * in_normal;

                    v_color = vec4( in_color, 1.0 );

                    v_camera = transpose(view)[2].xyz;
                    v_light_direction = -v_camera;
                } 
            """
    GEOMETRY = None
    FRAGMENT = """
                #version 330

                in vec3 v_normal;
                in vec3 v_pos;
                in vec4 v_color;
                in vec3 v_camera;
                in vec3 v_light_direction;

                out vec4 f_color;

                void main() {
                    vec3 normal = normalize(v_normal);
                    float inner_reflected = dot(reflect(normalize(v_pos-v_camera), normal), v_light_direction);
                    float inner_disperse = dot(normal, v_light_direction);
                    float intensity = 0.4+0.6*clamp(inner_disperse, 0.0, 1.0) + 0.3*clamp(inner_reflected, 0.0, 1.0);
                    f_color = vec4(intensity*v_color.xyz, v_color.t);
                }
            """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.ratio = 0.2
        self.length = 0.1
        self.location = (-0.85, -0.85)

        origin = np.array([[0.55, 0, 0], [0, 0.55, 0], [0, 0, 0.55]], dtype=np.float32)
        self._originBuffer = Float3DBuffer(mglctx=self._ctx, count=origin.shape[0])
        self._originBuffer.upload(origin)

        direction = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float32)
        self._directionBuffer = Float3DBuffer(
            mglctx=self._ctx, count=direction.shape[0]
        )
        self._directionBuffer.upload(direction)

        color = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float32)
        self._colorBuffer = Float3DBuffer(mglctx=self._ctx, count=color.shape[0])
        self._colorBuffer.upload(color)

        self.init_vao()

    def init_vao(self):
        self.mesh = ArrowMesh(mglctx=self._ctx, count=10)
        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self.mesh.pos.vbo, "3f", "in_pos"),
                (self.mesh.normal.vbo, "3f", "in_normal"),
                (self._directionBuffer.vbo, "3f/i", "in_dir"),
                (self._originBuffer.vbo, "3f/i", "in_origin"),
                (self._colorBuffer.vbo, "3f/i", "in_color"),
            ],
            self.mesh.idx.ibo,
        )

    def render(self, scales=None, modelview=None, projection=None):
        self._ctx.fbo.color_mask = False, False, False, False
        self._ctx.fbo.depth_mask = True
        self._ctx.clear()
        self._ctx.fbo.color_mask = True, True, True, True

        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())

        self.u_ratio.value = self.ratio
        self.u_length.value = self.length
        self.u_location.value = tuple(self.location)

        self._vao.render(mode=mgl.TRIANGLES, instances=3, vertices=-1, first=0)

    def list_properties(self):
        return []



# import struct
from typing import Callable
import moderngl as mgl
from PIL import Image, ImageOps

# from .buffer import Field3DBuffer
# from ..core import System
from .scene import Scene


class ImageContainer(object):
    def __init__(self, size=(1024, 1024), scene: Callable[[mgl.Context], Scene] = None):
        self._ctx = mgl.create_standalone_context()
        self._size = size
        self._color_rbo = self._ctx.renderbuffer(self._size)
        self._depth_rbo = self._ctx.depth_renderbuffer(self._size)
        self._fbo = self._ctx.framebuffer(self._color_rbo, self._depth_rbo)
        self._scene = None
        if not scene is None:
            self.attach_scene(scene)

    @property
    def scene(self):
        return self._scene

    def attach_scene(self, scene: Callable[[mgl.Context], Scene]):
        self._scene = scene(self._ctx)
        return self

    def upload(self, state):
        if self._scene is None:
            raise Exception("Scene should be attached first")
        self._scene.upload(state)
        return self

    def _render(self):
        self._fbo.use()
        self._scene.render(aspectratio=self._size[0] / self._size[1])

    def save(self, filename=None, crop=False):
        self._render()

        img = Image.frombytes("RGB", self._size, self._fbo.read())
        img = img.transpose(Image.FLIP_TOP_BOTTOM)

        if crop:
            inverted = ImageOps.invert(img)
            imageBox = inverted.getbbox()
            img = img.crop(imageBox)

        if filename is not None:
            img.save(filename)
            return self
        else:
            return img



import moderngl as mgl
import numpy as np

from .render import *
from .buffer import Float3DBuffer, IntBuffer
from ..core import NumericProperty, ListProperty, BooleanProperty


######################################################################################################

# GLSL functions for lattice points computation. For future usage.
#
# uniform ivec4 u_size; // Number of spins per each asis+number of atoms per unit cell.
# uniform mat3 u_generators; // Lattice generators.
# uniform vec3 u_representatives[]; // Atoms in the the unit cell.
#
# ivec4 reshape(int id) {
#     ivec4 stride;
#     stride.t = 1;
#     stride.z = stride.t*u_size.t;
#     stride.y = stride.z*u_size.z;
#     stride.x = stride.t*u_size.y;
#     return (id / stride) % u_size;
# }
#
# pos get_coordinates() {
#     ivec4 idx = reshape(gl_VertexID);
#     return u_generators*idx.xyz + u_representatives[idx.t];
# }


class IsolineRender(StateRender):
    VERTEX = """
            #version 330

            const float M_PI = 3.1415926535897932384626433832795;

            uniform vec3 u_camera; // camera position in world coordinates
            uniform float u_color_shift;
            uniform bool u_neglect_z; 
            uniform bool u_negate;

            in vec3 in_center; // center of atom
            in vec3 in_spin; // spin direction

            out VS_OUT {
                vec3 origin; // atom position in world coordiantes
                vec3 spin; // spin direction in spin space
                vec3 light_direction; 
                vec3 color;
            } vs_out;

            vec3 hsv2rgb(vec3 c) {
                vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
                vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
                return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
            }

            vec3 spin_to_color(vec3 spin) {
                float s = spin.z;
                float phi = (atan(spin.y, spin.x)/M_PI+1)/2;;
                vec3 hsv = vec3( phi, clamp(1-0.9*s, 0, 1), clamp(1+0.6*s, 0, 1) );
                hsv.x += u_color_shift;                        
                return hsv2rgb(hsv);
            }

            void main() {
                vec3 spin = normalize(in_spin);
                if(u_negate) spin=-spin;
                vs_out.origin = in_center;
                vs_out.spin = spin;
                vs_out.light_direction = - normalize(vs_out.origin-u_camera);
                if (u_neglect_z) {
                    vs_out.color = spin_to_color(vec3(spin.xy, 0));
                } else {
                    vs_out.color = spin_to_color(spin);
                };
            } 
        """
    GEOMETRY = """
            #version 330 core

            layout (lines_adjacency) in;
            layout (triangle_strip, max_vertices = 8) out;

            uniform mat4 u_view;
            uniform mat4 u_projection;
            uniform float u_width=0.0;

            uniform bool u_inverted;
            uniform vec3 u_axis; // Direction of the vector field on the isoline.


            in VS_OUT {
                vec3 origin; // atom position in world coordiantes
                vec3 spin; // spin direction in spin space
                vec3 light_direction; 
                vec3 color;
            } vs_in[];

            out vec3 v_position;
            out vec3 v_color;
            out vec3 v_light_direction;
            // out vec3 v_normal;

            int idx;
            struct {
                vec3 position;
                vec3 color;
                vec3 light_direction;
                vec4 glposition;
            } data[4];

            void run(int a, int b, int c) {
                // Compute barycentric coordinates q of u_axis in triangle in spanned by three spins a,b,c.
                // q.x+q.y+q.z=1, and a*q.x+b*q.y+c*q.z=q.t*u_axis
                mat4 m = mat4( 
                    vec4( vs_in[a].spin, 1.0 ),
                    vec4( vs_in[b].spin, 1.0 ),
                    vec4( vs_in[c].spin, 1.0 ),
                    vec4( -u_axis, 0.0)
                );
                vec4 q = inverse(m) * vec4(0.0,0.0,0.0,1.0) ;
                if(q.w<0) q=-q;
                // If u_axis is inside the spherical triangle spanned by a,b,c, then add the point to the curve.
                if(q.x>0 && q.y>0 && q.z>0 && q.w>0) {
                    data[idx].position = q.x*vs_in[a].origin + q.y*vs_in[b].origin + q.z*vs_in[c].origin;
                    data[idx].color = q.x*vs_in[a].color + q.y*vs_in[b].color + q.z*vs_in[c].color;
                    data[idx].light_direction = q.x*vs_in[a].light_direction + q.y*vs_in[b].light_direction + q.z*vs_in[c].light_direction;
                    data[idx].glposition = u_projection * u_view * vec4(data[idx].position, 1.0);
                    idx += 1;
                };
            }

            void main() {
                idx = 0;
                // Find intersections with the simplex.
                // v_color = random_vec();
                run(0, 1, 2);
                run(1, 2, 3);
                run(2, 3, 0);
                run(3, 0, 1);      

                // Generate vertices.
                for(int n=0; n<idx; n++) {
                    // Compute normal to the line.
                    vec2 tangent=vec2(0.,0.);
                    if(n>0) {
                        tangent += normalize(data[n].glposition.xy - data[n-1].glposition.xy);
                    };
                    if(n<idx-1) {
                        tangent += normalize(data[n+1].glposition.xy - data[n].glposition.xy);
                    };
                    tangent = normalize(tangent);
                    vec2 normal = vec2(-tangent.y,tangent.x); 

                    // Emit vertices.
                    v_position = data[n].position;
                    v_color = data[n].color;
                    v_light_direction = data[n].light_direction;
                    vec2 xy = data[n].glposition.xy;
                    if(n==0) xy-=tangent*u_width/4;
                    if(n==idx-1) xy+=tangent*u_width/4;
                    gl_Position = vec4(xy+normal*u_width, data[n].glposition.z, 1.0);
                    EmitVertex();                    

                    gl_Position = vec4(xy-normal*u_width, data[n].glposition.z, 1.0);
                    EmitVertex();                    
                };
                EndPrimitive();
            }   
        """
    FRAGMENT = """
            #version 330

            uniform vec3 u_camera;
            uniform bool u_lighting;

            // in vec3 v_normal;
            in vec3 v_position;
            in vec3 v_color;
            in vec3 v_light_direction;

            out vec4 f_color;

            void main() {
                if (u_lighting) {
                    vec3 normal = normalize(cross(dFdx(v_position), dFdy(v_position)));
                    // vec3 normal = normalize(v_normal);
                    float inner_reflected = dot(reflect(normalize(v_position-u_camera), normal), v_light_direction);
                    float inner_disperse = dot(normal, v_light_direction);
                    float intensity = 0.4+0.6*clamp(inner_disperse, 0.0, 1.0) + 0.3*clamp(inner_reflected, 0.0, 1.0);
                    f_color = vec4(intensity*v_color.xyz, 1.0);
                    // f_color = vec4(abs(normal), 1.0); // DEBUG
                } else {
                     f_color = vec4(v_color.xyz, 1.0);
                };
            }

        """

    def __init__(
        self,
        coordinatesBuffer: Float3DBuffer = None,
        maskBuffer: Float3DBuffer = None,
        indexBuffer: IntBuffer = None,
        scalarMaskBuffer=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        assert self._system.dim == 3

        self._coordinatesBuffer = coordinatesBuffer
        self._coordinatesBuffer.upload(self._system.spin_positions())

        self._idx = indexBuffer

        self.init_vao()

        self.lighting = False
        self.inverted = True
        self.negate = False
        self.color_shift = 0.81
        self.axis = (0, 0, 1)
        self.neglect_z = False
        self.width = 0.01

    def init_vao(self):
        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self._state.vbo, "3f/v", "in_spin"),
                (self._coordinatesBuffer.vbo, "3f/v", "in_center"),
            ],
            self._idx.ibo,
        )

    def render(
        self, scales=None, modelview=None, projection=None, camera_position=None
    ):
        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())
        self.u_camera.value = tuple(camera_position)

        self.u_color_shift.value = self.color_shift
        self.u_lighting.value = self.lighting
        self.u_negate.value = self.negate
        self.u_axis.value = self.axis
        self.u_neglect_z.value = self.neglect_z
        # self.u_inverted.value = self.inverted
        self.u_width.value = self.width

        self._ctx.enable_only(mgl.DEPTH_TEST)

        self._ctx.line_width = 2
        self._vao.render(mode=mgl.LINES_ADJACENCY, vertices=-1, first=0)

    def list_properties(self):
        return [
            BooleanProperty(self, "lighting", title="Lighting"),
            BooleanProperty(self, "inverted", title="Invert colors"),
            BooleanProperty(self, "negate", title="Negate spin"),
            BooleanProperty(self, "neglect_z", title="Ignore z-projection"),
            NumericProperty(
                self, "color_shift", title="Hue", min=0.0, max=1.0, count=360
            ),
            NumericProperty(
                self, "width", title="Width", min=0.0, max=0.02, count=1000
            ),
        ]


####################################################################################################################


class IsolinesRender(IsolineRender):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.npetals = 16
        self.ntheta = 1
        self.theta_shift = 0
        self.phi_shift = 0

    def render(
        self, scales=None, modelview=None, projection=None, camera_position=None
    ):
        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())
        self.u_camera.value = tuple(camera_position)

        # self.u_inverted.value = self.inverted
        self.u_color_shift.value = self.color_shift
        self.u_lighting.value = self.lighting
        self.u_neglect_z.value = self.neglect_z
        self.u_negate.value = self.negate
        self.u_width.value = self.width
        # self._ctx.enable_only(mgl.DEPTH_TEST)
        thetas = np.linspace(0, np.pi / 2, int(np.round(self.ntheta)), endpoint=False)
        theta_offset = np.pi / 2 / len(thetas) * self.theta_shift
        for theta in thetas:
            z, xy = np.sin(theta + theta_offset), np.cos(theta + theta_offset)
            k = int(np.round(np.cos(theta) * self.npetals)) + 1
            phi_offset = 2 * np.pi / k * self.phi_shift
            for phi in np.linspace(0, 2 * np.pi, k, endpoint=False) + phi_offset:
                x, y = xy * np.cos(phi), xy * np.sin(phi)
                self.u_axis.value = (x, y, z)
                self._vao.render(mode=mgl.LINES_ADJACENCY, vertices=-1, first=0)
        for theta in -thetas[1:]:
            z, xy = np.sin(theta + theta_offset), np.cos(theta + theta_offset)
            k = int(np.round(np.cos(theta) * self.npetals)) + 1
            phi_offset = 2 * np.pi / k * self.phi_shift
            for phi in np.linspace(0, 2 * np.pi, k, endpoint=False) + phi_offset:
                x, y = xy * np.cos(phi), xy * np.sin(phi)
                self.u_axis.value = (x, y, z)
                self._vao.render(mode=mgl.LINES_ADJACENCY, vertices=-1, first=0)

    def list_properties(self):
        return [
            BooleanProperty(self, "lighting", title="Lighting"),
            BooleanProperty(self, "inverted", title="Invert colors"),
            BooleanProperty(self, "negate", title="Negate spin"),
            BooleanProperty(self, "neglect_z", title="Ignore z-projection"),
            NumericProperty(
                self,
                "phi_shift",
                title="Polar angle offset",
                min=-1.0,
                max=1.0,
                count=1000,
            ),
            NumericProperty(
                self,
                "theta_shift",
                title="Azimuthal angle offset",
                min=-1.0,
                max=1.0,
                count=1000,
            ),
            NumericProperty(
                self, "npetals", title="Polar angle substeps", min=1, max=10, count=10
            ),
            NumericProperty(
                self, "ntheta", title="Azimuthal angle substeps", min=1, max=5, count=5
            ),
            NumericProperty(
                self, "color_shift", title="Hue", min=0.0, max=1.0, count=360
            ),
            NumericProperty(
                self, "width", title="Width", min=0.0, max=0.02, count=1000
            ),
        ]



import numpy as np
import taichi as ti
from scipy.interpolate import interp1d


######################################################################################
# Aux functions.

def simpson(x:np.ndarray) -> np.ndarray:
    if x.shape[0]==1:
        return x[0]
    assert x.shape[0]%2==1
    return np.sum(x[:-2:2] + 4*x[1:-1:2] + x[2::2], axis=0)/6

######################################################################################
# Functions of wavelength.

class Function:
    def __call__(self, x):
        x = np.asarray(x)
        if x.ndim==0:
            x = x[None]
        assert x.ndim == 1
        return self.compute(x)

    def compute(self, x: np.ndarray) -> np.ndarray:
        raise NotImplementedError
    
    def __call__(self, x: np.ndarray) -> np.ndarray:
        return self.compute(x)

class ConstantFunction:
    def __init__(self, x):
        self.x = np.asarray(x)

    def compute(self, x: np.ndarray) -> np.ndarray:
        return np.repeat(self.x[None], x.shape[0], axis=0)

class GaussianFunction(Function):
    def __init__(self, center, width):
        self._center = np.asarray(center)
        self._width = np.asarray(width)
        self._2width2 = 2*self._width**2

    def compute(self, x):
        return np.exp(-(x[:,None]-self._center[None])**2/self._2width2)/self._width

class SimplifiedRGB(GaussianFunction):
    def __init__(self):
        super().__init__(
            center=[0.610,0.540,0.450],
            width=[0.050,0.040,0.030],
            )

class BlackbodyRadiation(Function):
    h = 6.62607015e-34 # J⋅Hz−1
    c = 299792458 # m/s
    k = 1.380649e-23 # J⋅K−1

    def __init__(self, temperature_in_K=[5800]):
        self.T = np.asarray(temperature_in_K)
        assert self.T.ndim==1, f"{self.T.shape=}"

    def compute(self, x):
        assert x.ndim==1, f"{x.shape=}"
        wavelength = x*1e-6 # meters
        b = 2.897771955e-3 # [m K], Wien displacement constant.
        print(f"Radiation peak at {b/self.T[0]*1e6:.3f} um")
        result = (2*self.h*self.c**2/wavelength[:,None]**5)/( np.exp( self.h*self.c/(self.k*self.T[None,:]*wavelength[:,None]) ) - 1 )
        return result 

class FunctionTable(Function):
    def __init__(self, x, y):
        self.x = np.asarray(x)
        self.y = np.asarray(y)
        self.poly = interp1d(x=x, y=y, axis=0, bounds_error=False, fill_value=0)

    @classmethod
    def from_csv(cls, name:str) -> 'FunctionTable':
        data = np.loadtxt(fname=name, dtype=np.float32, delimiter=None)
        return cls(x=data[:,0], y=data[:,1:])

    def compute(self, x):
        return self.poly(x)


################################################################################################################################
# Multispectral cross-polarization.

@ti.kernel
# cuda.jit((nbfloat[:,:,:,::1], nbfloat, nbfloat[::1], nbfloat[::1], nbfloat[::1], nbfloat, nbfloat, nbfloat, nbfloat, nbfloat[:,:,::1]))
def cross_polarization_filter_multi(n:ti.template(), dz:float, lambdas:ti.template(), nes:ti.template(), nos:ti.template(), 
                                    polarizerx:float, polarizery:float, filtrx:float, filtry:float, out:ti.template()):
    _sx, _sy, sz, _ = n.shape
    for w,x,y in out:
        # Constants for given wavelength
        dzoverlambda = dz/lambdas[w]
        ne = nes[w]
        no = nos[w]
        nooverne2 = (no/ne)**2
        # Incident wave field.
        Ex = ti.math.vec2(polarizerx,0)
        Ey = ti.math.vec2(polarizery,0)
        # Go layer by layer.
        for z in range(sz):
            mx = n[x,y,z,0]
            my = n[x,y,z,1]
            mz = n[x,y,z,2]
            ll = ti.math.sqrt(mx**2+my**2)+1e-15
            nx = mx/ll
            ny = my/ll
            p = Ex*nx+Ey*ny
            neff = no/ti.math.sqrt( nooverne2*(mx**2+my**2)+mz**2 )
            gammaz = np.pi*dzoverlambda*(neff-no)
            factor = ti.math.cmul(p, ti.math.vec2(1-ti.math.cos(2*gammaz), ti.math.sin(2*gammaz)))
            Ex = Ex-factor*nx
            Ey = Ey-factor*ny
        # Apply filter.
        out[w,x,y] = ti.math.length(Ex*filtrx+Ey*filtry)


def render_cp_multi(nn:np.ndarray, wavelengths:np.ndarray, ne:Function, no:Function, emission:Function, efficiency:Function, 
                    thickness:float, polarizer:float=0., deltafilter:float=np.pi/2):
    # assert isinstance(x, State)
    def todev(x):
        if isinstance(x, np.ndarray):
            f = ti.field(ti.f32, shape=x.shape)
            f.from_numpy(x.astype(np.float32))
            return f
        raise NotImplementedError

    assert nn.ndim==4 and nn.shape[3]==3, f"{nn.shape=}"
    f_nn = todev(nn)
    #
    assert wavelengths.ndim==1, f"{wavelengths.shape=}"
    f_wavelengths = todev(wavelengths)
    #
    n_ne = ne.compute(wavelengths)
    assert n_ne.shape == wavelengths.shape, f"{n_ne.shape=} {wavelengths.shape=}"
    f_ne = todev(n_ne)
    #
    n_no = no.compute(wavelengths)
    assert n_no.shape == wavelengths.shape, f"{n_no.shape=} {wavelengths.shape=}"
    f_no = todev(n_no)
    #
    n_emission = emission.compute(wavelengths)
    assert n_emission.shape == wavelengths.shape+(1,), f"{n_emission.shape=} {wavelengths.shape=}"
    n_efficiency = efficiency.compute(wavelengths)
    assert n_efficiency.shape == wavelengths.shape+(3,), f"{n_efficiency.shape=} {wavelengths.shape=}"
    # Allocate storage
    f_output = ti.field(ti.f32, shape=wavelengths.shape+nn.shape[:2])
    # Compute parameters.
    sz = nn.shape[2]
    dz = thickness/(sz-1)
    polarizer_vec=np.array([ np.cos(polarizer), np.sin(polarizer) ])
    filter_vec=np.array([ np.cos(polarizer+deltafilter), np.sin(polarizer+deltafilter) ])
    # Compute amplitudes for each wavelength.
    cross_polarization_filter_multi(f_nn, dz, f_wavelengths, f_ne, f_no, polarizer_vec[0], polarizer_vec[1], filter_vec[0], filter_vec[1], f_output)
    n_output = f_output.to_numpy()
    # Project to RGB.
    n_amp = n_output*n_emission[:,:,None] # (lambda,y,x)
    n_rgb = simpson(n_amp[:,:,:,None]*n_efficiency[:,None,None,:]) # (y,x,rgb)
    n_rgb /= np.max(n_rgb)
    return np.clip(n_rgb, 0., 1.) # Save in linear color space.



import moderngl as mgl
import numpy as np

from .render import StateRender
from .buffer import Float3DBuffer, IntBuffer
from ..core import NumericProperty, ListProperty, BooleanProperty
from .solid import ConicMesh, CylinderMesh, DoubleConeMesh, ArrowMesh


######################################################################################################


class VectorRender(StateRender):
    VERTEX = """
                #version 330

                const float M_PI = 3.1415926535897932384626433832795;

                uniform mat4 u_view;
                uniform mat4 u_projection;
                uniform bool u_negate;

                uniform vec3 u_camera; // camera position in world coordinates
                // uniform vec3 u_light_direction; // directed light source in world coordinates
                uniform bool u_use_mask; // false to display all spins
                uniform bool u_use_scalar_mask; // true to use in_scalar

                uniform float u_min_size;
                uniform float u_max_size;
                uniform float u_power_size;
                uniform float u_min_threshold;
                uniform ivec3 u_size;
                uniform ivec3 u_clip_min;
                uniform ivec3 u_clip_max; 
                uniform bool u_clip_invert; 
                uniform float u_min_scalar;
                uniform float u_max_scalar;
                uniform float u_color_shift;
                uniform bool u_director_mode;
                uniform float u_width;
                uniform float u_saturation_mag;

                in vec3 in_center; // center of atom
                in vec3 in_pos; // mesh point in mesh coordinates
                in vec3 in_normal; // normal to mesh at in_pos in mesh coordinates
                in vec3 in_spin; // spin direction
                in vec3 in_mask; // invisible direction
                in float in_scalar; // invisible direction

                out vec3 v_pos; // vertex position in world coordiantes
                out vec3 v_normal; // normal to surface in world coordinates
                out vec4 v_color;

                out vec3 v_camera;
                out vec3 v_light_direction;

                mat3 rotmat(vec3 z) {
                    vec3 a = cross(z, vec3(1.0,0.0,0.0));
                    vec3 b = cross(z, vec3(0.0,1.0,0.0));
                    vec3 c = (length(a)>length(b)) ? a : b;
                    vec3 x = normalize(c);
                    vec3 y = cross(x, z);
                    return mat3(x, y, z);
                }

                vec3 hsv2rgb(vec3 c) {
                    vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
                    vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
                    return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
                }

                vec3 spin_to_color(vec3 spin) {
                    if(u_director_mode) {
                        vec3 pr = spin.yzx*spin.zxy;
                        // pr /= max( max(abs(pr.x),abs(pr.y)), abs(pr.z));
                        pr /= length(pr);
                        vec3 rgb = 0.5-0.5*pr;
                        return rgb;
                    } else {
                        float s = spin.z;
                        if (u_saturation_mag!=1.0) {
                            // s = 1+(s-1)*u_saturation_mag;
                            s = cos(u_saturation_mag*acos(s));
                        };
                        float phi = (atan(spin.y, spin.x)/M_PI+1)/2;
                        vec3 hsv = vec3( phi, clamp(1-s, 0, 1), clamp(1+s, 0, 1) );
                        hsv.x += u_color_shift;                        
                        return hsv2rgb(hsv);
                    };
                }
 
                void main() {
                    float len = length(in_spin);
                    vec3 spin = u_negate?-in_spin/len:in_spin/len;
                    float scale;
                    if(u_use_mask) {
                        vec3 mask = u_negate?-in_mask:in_mask;
                        scale = u_use_scalar_mask ? (in_scalar-u_min_scalar) / (u_max_scalar-u_min_scalar) : (1 - dot(mask, spin)) / 2;
                    } else {
                        scale = 1.0;
                    };
                    scale *= len;
                    if (scale<u_min_threshold) { 
                        scale = 0;
                    } else {
                        scale = u_min_size + pow(scale, u_power_size) * (u_max_size-u_min_size);
                    };

                    int i = gl_InstanceID; int iz = i % u_size.z;
                    i /= u_size.z; int iy = i % u_size.y;
                    i /= u_size.y; int ix = i % u_size.x;
                    bool inbox = (ix<u_clip_min.x || ix>u_clip_max.x 
                        || iy<u_clip_min.y || iy>u_clip_max.y 
                        || iz<u_clip_min.z || iz>u_clip_max.z);
                    if (inbox != u_clip_invert) scale = 0;

                    mat3 mesh = rotmat(spin);
                    vec3 pos = scale*in_pos;
                    pos.xy *= u_width;
                    v_pos = in_center + mesh * pos;
                    gl_Position = u_projection * u_view * vec4(v_pos, 1.0);
                    v_normal = mesh * in_normal;

                    v_camera = u_camera; // * mat3(u_view);
                    // vec3 light_direction = u_light_direction;
                    // vec3 light_direction = - u_camera + v_pos;
                    // v_light_direction = normalize(light_direction * mat3(u_view));
                    v_light_direction = - normalize(v_pos-v_camera);

                    v_color = vec4( spin_to_color(spin), 1.0 );
                } 
            """
    GEOMETRY = None
    FRAGMENT = """
                #version 330

                uniform bool u_lighting;

                in vec3 v_normal;
                in vec3 v_pos;
                in vec4 v_color;

                in vec3 v_camera;
                in vec3 v_light_direction;

                out vec4 f_color;

                void main() {
                    if (u_lighting) {
                        vec3 normal = normalize(v_normal);
                        float inner_reflected = dot(reflect(normalize(v_pos-v_camera), normal), v_light_direction);
                        float inner_disperse = dot(normal, v_light_direction);
                        float intensity = 0.4+0.6*clamp(inner_disperse, 0.0, 1.0) + 0.3*clamp(inner_reflected, 0.0, 1.0);
                        f_color = vec4(intensity*v_color.xyz, v_color.t);
                    } else {
                        f_color = v_color;
                    };
                }
            """

    def __init__(
        self,
        _light=(1, 2, -3),
        coordinatesBuffer=None,
        maskBuffer=None,
        scalarMaskBuffer=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self._coordinatesBuffer = coordinatesBuffer
        self._coordinatesBuffer.upload(self._system.spin_positions())

        self._maskBuffer = maskBuffer
        self._scalarMaskBuffer = scalarMaskBuffer

        # self.mesh = ConicMesh(mglctx=self._ctx)

        self._detalization = 1
        self._mesh_shape = "Cone"
        self.init_vao()

        self.use_mask = False
        self.use_scalar_mask = False
        self.director_mode = False
        self.base_size = 1.0
        self.min_size = 0.0
        self.max_size = 1.2
        self.min_scalar = 0.0
        self.max_scalar = 1.0
        self.power_size = 1.0
        self.min_threshold = 0.75
        assert self._system.dim == 3
        self.size = self._system.size
        self.clip_min = (0, 0, 0)
        self.clip_max = tuple(self.size)
        self.clip_invert = False
        self.color_shift = 0.81
        self.width = 0.75
        self.saturation_mag = 1.0
        self.lighting = True
        self.negate = False

    def init_vao(self):
        self.mesh = self.MESH_NAMES[self._mesh_shape](
            mglctx=self._ctx, count=int(10 * self._detalization)
        )
        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self.mesh.pos.vbo, "3f", "in_pos"),
                (self.mesh.normal.vbo, "3f", "in_normal"),
                (self._state.vbo, "3f/i", "in_spin"),
                (self._coordinatesBuffer.vbo, "3f/i", "in_center"),
                (self._maskBuffer.vbo, "3f/i", "in_mask"),
                (self._scalarMaskBuffer.vbo, "f/i", "in_scalar"),
            ],
            self.mesh.idx.ibo,
        )

    MESH_NAMES = {
        "Cone": ConicMesh,
        "Double cone": DoubleConeMesh,
        "Cylinder": CylinderMesh,
        "Arrow": ArrowMesh,
    }

    @property
    def mesh_shape(self):
        for n, t in self.MESH_NAMES.items():
            if isinstance(self.mesh, t):
                return n
        raise NotImplementedError

    @mesh_shape.setter
    def mesh_shape(self, name):
        self._mesh_shape = name
        self.init_vao()

    @property
    def detalization(self):
        return self._detalization

    @detalization.setter
    def detalization(self, value):
        print(f"{self._detalization = },\t {value = }")
        self._detalization = value
        self.init_vao()

    def render(
        self, scales=None, modelview=None, projection=None, camera_position=None
    ):
        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())
        self.u_camera.value = tuple(camera_position)
        # self.u_light_direction.value = tuple(self._light)
        self.u_use_mask.value = self.use_mask
        self.u_min_size.value = self.min_size * self.base_size
        self.u_max_size.value = self.max_size * self.base_size
        self.u_use_scalar_mask.value = self.use_scalar_mask
        self.u_min_scalar.value = self.min_scalar
        self.u_max_scalar.value = self.max_scalar
        self.u_power_size.value = self.power_size
        self.u_min_threshold.value = self.min_threshold
        self.u_size.value = tuple(self.size)
        self.u_clip_min.value = tuple(np.array(self.clip_min, dtype=np.int32))
        self.u_clip_max.value = tuple(np.array(self.clip_max, dtype=np.int32))
        self.u_clip_invert.value = self.clip_invert
        self.u_color_shift.value = self.color_shift
        self.u_director_mode.value = self.director_mode
        self.u_width.value = self.width
        self.u_saturation_mag.value = self.saturation_mag
        self.u_lighting.value = self.lighting
        self.u_negate.value = self.negate

        self._vao.render(
            mode=mgl.TRIANGLES,
            instances=self._system.number_of_spins,
            vertices=-1,
            first=0,
        )

        # print(self.u_min_scalar.value, self.u_max_scalar.value, self.u_use_scalar_mask.value, self.u_use_mask.value,)

    def list_properties(self):
        return [
            BooleanProperty(self, "lighting", title="Lighting"),
            ListProperty(
                self, "mesh_shape", title="Shape", values=sorted(self.MESH_NAMES.keys())
            ),
            BooleanProperty(self, "director_mode", title="Director mode"),
            BooleanProperty(self, "negate", title="Negate spin"),
            NumericProperty(self, "width", title="Width", min=0.01, max=2.0, count=100),
            NumericProperty(
                self, "color_shift", title="Hue", min=0.0, max=1.0, count=360
            ),
            NumericProperty(
                self, "saturation_mag", title="Saturation", min=1.0, max=10.0, count=500
            ),
            NumericProperty(
                self, "min_size", title="Min. size", min=0.0, max=2.0, count=30
            ),
            NumericProperty(
                self, "max_size", title="Max. size", min=0.0, max=2.0, count=30
            ),
            NumericProperty(
                self, "power_size", title="Steepness", min=0.0, max=2.0, count=30
            ),
            NumericProperty(
                self,
                "min_threshold",
                title="Min. threshold",
                min=0.0,
                max=1.0,
                count=500,
            ),
            NumericProperty(
                self, "detalization", title="Detalization", min=1.0, max=10.0, count=10
            ),
        ]



from .core import log, System

__all__ = ["log", "System"]

#########################################################################################################################
# Load plugins

import importlib
from importlib.util import find_spec
from importlib.metadata import entry_points
import sys


class LazyModule:
    def __init__(self, name: str, extra: str):
        self.name = name
        self.extra = extra
        self.module = None

    def __getattr__(self, attr):
        if self.module is None:
            log.debug(f"Importing plugin {self.name}")
            s = find_spec(self.name)
            if s:
                self.module = importlib.import_module(self.name)
            else:
                raise ImportError(
                    f"Plugin {self.name} is not found. Install the library with `uv sync --extra {self.extra}` or `uv sync --all-extras`"
                )
        return getattr(self.module, attr)


def load_plugins():
    plugins = {}
    for ep in entry_points(group="my_library_plugins"):
        module = LazyModule(ep.value, ep.name)
        setattr(sys.modules[__name__], ep.name, module)
    return plugins


load_plugins()



import sadcompressor as sad
import numpy as np

from ..core import System


def load_lcsim_sad_single(
    reader: sad.ContainerReader,
) -> tuple[np.ndarray, System, dict]:
    x, y, z = list(reader[n] for n in ["x", "y", "z"])
    director = np.stack([x, y, z], axis=-1).transpose((2, 1, 0, 3))[:, :, :, None]
    settings = reader["s"]
    system = System.cubic(size=director.shape[:3])
    return director, system, settings



import moderngl as mgl
import numpy as np

from .render import *
from .buffer import Float3DBuffer, IntBuffer
from ..core import NumericProperty, ListProperty, BooleanProperty


######################################################################################################


class LevelRender(StateRender):
    VERTEX = """
            #version 330

            in vec3 in_center; // center of atom
            in vec3 in_spin; // spin direction

            out VS_OUT {
                vec3 origin; // atom position in world coordiantes
                vec3 spin; // spin direction in spin space
            } vs_out;

            void main() {
                vec3 spin = normalize(in_spin);
                vs_out.origin = in_center;
                vs_out.spin = spin;
            } 
        """
    GEOMETRY = """
            #version 330 core

            const float M_PI = 3.1415926535897932384626433832795;


            layout (lines_adjacency) in;
            layout (triangle_strip, max_vertices = 16) out;

            uniform vec3 u_axis; // Zenith direction.
            uniform float u_level;
            uniform bool u_constant_theta;


            uniform mat4 u_view;
            uniform mat4 u_projection;
            uniform int u_subdivide=1;

            uniform vec3 u_camera; // camera position in world coordinates
            uniform float u_color_shift;
            uniform bool u_neglect_z;
            uniform bool u_inverted;

            in VS_OUT {
                vec3 origin;
                vec3 spin;
            } vs_in[];

            out vec3 v_position;
            out vec3 v_color;
            out vec3 v_light_direction;

            struct Input {
                vec3 origin;
                vec3 spin;
                float potential;
            };

            vec3 hsv2rgb(vec3 c) {
                vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
                vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
                return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
            }

            vec3 spin_to_color(vec3 spin) {
                float s = spin.z;
                float phi = (atan(spin.y, spin.x)/M_PI+1)/2;
                vec3 hsv = vec3( phi, clamp(1-0.9*s, 0, 1), clamp(1+0.6*s, 0, 1) );
                hsv.x += u_color_shift;                        
                return hsv2rgb(hsv);
            }

            void run0(Input a, Input b) {
                // f(t)=f(0)+t*(f(1)-f(0))=0
                // t = f(0)/(f(0)-f(1)))
                float t = a.potential/(a.potential-b.potential);
                if (!(0<t && t<1)) return;
                v_position = mix(a.origin, b.origin, t);
                vec3 spin = mix(a.spin, b.spin, t); 
                if (u_neglect_z) {
                    v_color = spin_to_color(vec3(spin.xy, 0));
                } else {
                    v_color = spin_to_color(spin);
                };

                v_light_direction = - normalize(mix(a.origin, b.origin, t)-u_camera);

                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
            }

            void run(Input a, Input b, Input c, Input d) {
                // There should be no point on the edge ab.
                run0(a, c);
                run0(b, c);
                run0(a, d);
                run0(b, d);
                run0(c, d);
            }

            void process(Input a0, Input a1, Input a2, Input a3) {
                     if (sign(a0.potential)==sign(a1.potential)) run(a0, a1, a2, a3);
                else if (sign(a0.potential)==sign(a2.potential)) run(a0, a2, a1, a3);
                else if (sign(a1.potential)==sign(a2.potential)) run(a1, a2, a3, a0);
                EndPrimitive();
            }

            void interpolate(in float t0, in float t1, in float t2, in float t3, out Input res) {
                res.origin = t0*vs_in[0].origin+t1*vs_in[1].origin+t2*vs_in[2].origin+t3*vs_in[3].origin;
                res.spin = normalize(t0*vs_in[0].spin+t1*vs_in[1].spin+t2*vs_in[2].spin+t3*vs_in[3].spin);
                if (u_constant_theta) {
                    // float theta = acos(dot(res.spin, u_axis));
                    // res.potential = 2*theta/M_PI - u_level - 1.0;

                    // res.potential = cos(theta-0.5*M_PI*u_level);
                
                    float theta = dot(res.spin, u_axis);
                    res.potential = theta - u_level;
                } else {
                    float phi = atan(res.spin.y, res.spin.x);
                    res.potential = cos(phi-M_PI*u_level);
                };
            }

            void main() {
                Input a0, a1, a2, a3;
                interpolate(1, 0, 0, 0, a0);
                interpolate(0, 1, 0, 0, a1);
                interpolate(0, 0, 1, 0, a2);
                interpolate(0, 0, 0, 1, a3);
                process(a0, a1, a2, a3);

                
                Input ac;
                interpolate(0.25, 0.25, 0.25, 0.25, ac);
                process(a0, a1, a2, ac);
                process(a1, a2, a3, ac);
                process(a2, a3, a0, ac);
                process(a3, a0, a1, ac);
                

                /*
                Input a01, a02, a03, a12, a13, a23;
                interpolate(0.5, 0.5, 0, 0, a01);
                interpolate(0.5, 0, 0.5, 0, a02);
                interpolate(0.5, 0, 0, 0.5, a03);
                interpolate(0, 0.5, 0.5, 0, a12);
                interpolate(0, 0.5, 0, 0.5, a13);
                interpolate(0, 0, 0.5, 0.5, a23);
                process(a0, a01, a02, a03);
                process(a1, a01, a12, a13);
                process(a2, a02, a12, a23);
                process(a3, a03, a13, a23);
                process(a01, a02, a12, ac);
                process(a01, a03, a13, ac);
                process(a02, a03, a23, ac);
                process(a12, a13, a23, ac);
                */
            }   
        """
    FRAGMENT = """
            #version 330

            uniform vec3 u_camera;
            uniform bool u_lighting;

            // in vec3 v_normal;
            in vec3 v_position;
            in vec3 v_color;
            in vec3 v_light_direction;

            out vec4 f_color;

            void main() {
                if (u_lighting) {
                    vec3 normal = normalize(cross(dFdx(v_position), dFdy(v_position)));
                    // vec3 normal = normalize(v_normal);
                    float inner_reflected = dot(reflect(normalize(v_position-u_camera), normal), v_light_direction);
                    float inner_disperse = dot(normal, v_light_direction);
                    float intensity = 0.4+0.6*clamp(inner_disperse, 0.0, 1.0) + 0.3*clamp(inner_reflected, 0.0, 1.0);
                    f_color = vec4(intensity*v_color.xyz, 1.0);
                    // f_color = vec4(abs(normal), 1.0); // DEBUG
                } else {
                     f_color = vec4(v_color.xyz, 1.0);
                };
            }

        """

    ANGLENAMES = ["polar angle", "azimuthal angle"]

    def __init__(
        self,
        coordinatesBuffer=None,
        maskBuffer=None,
        indexBuffer=None,
        scalarMaskBuffer=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        assert self._system.dim == 3

        self.u_view = self._prog["u_view"]
        self.u_projection = self._prog["u_projection"]
        self.u_color_shift = self._prog["u_color_shift"]
        self.u_level = self._prog["u_level"]
        self.u_constant_theta = self._prog["u_constant_theta"]
        self.u_lighting = self._prog["u_lighting"]
        self.u_axis = self._prog["u_axis"]
        self.u_neglect_z = self._prog["u_neglect_z"]

        self._coordinatesBuffer = coordinatesBuffer
        self._coordinatesBuffer.upload(self._system.spin_positions())

        self._idx = indexBuffer

        self.init_vao()

        self.lighting = True
        self.inverted = True
        self.color_shift = 0.81
        self.level = 0.0
        self.constant_theta = True
        self.axis = (0, 0, 1)

    def init_vao(self):
        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self._state.vbo, "3f/v", "in_spin"),
                (self._coordinatesBuffer.vbo, "3f/v", "in_center"),
            ],
            self._idx.ibo,
        )

    @property
    def angle_name(self):
        return self.ANGLENAMES[0 if self.constant_theta else 1]

    @angle_name.setter
    def angle_name(self, name):
        self.constant_theta = name == self.ANGLENAMES[0]
        # self.init_vao()

    def render(
        self, scales=None, modelview=None, projection=None, camera_position=None
    ):
        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())
        self.u_camera.value = tuple(camera_position)

        self.u_axis.value = self.axis
        # self.u_inverted.value = self.inverted
        self.u_constant_theta.value = self.constant_theta
        self.u_neglect_z.value = self.constant_theta
        self.u_color_shift.value = self.color_shift
        self.u_level.value = self.level
        self.u_lighting.value = self.lighting

        self._ctx.enable_only(mgl.DEPTH_TEST)

        self._vao.render(mode=mgl.LINES_ADJACENCY, vertices=-1, first=0)

    def list_properties(self):
        return [
            BooleanProperty(self, "lighting", title="Lighting"),
            BooleanProperty(self, "inverted", title="Invert colors"),
            ListProperty(self, "angle_name", title="Constant", values=self.ANGLENAMES),
            # BooleanProperty(self, "constant_theta", title="Constant z-projection of spin"),
            NumericProperty(
                self, "level", title="Level", min=-1.0, max=1.0, count=1000
            ),
            NumericProperty(
                self, "color_shift", title="Hue", min=0.0, max=1.0, count=360
            ),
        ]


####################################################################################################################



import numpy as np
import json

from .log import log
from .system import System


def load_magnes_npz(filename, path="PATH") -> tuple[np.ndarray, System]:
    raise NotImplementedError


def load_lcsim_npz(filename, path="PATH") -> tuple[np.ndarray, System, dict]:
    log.debug(f"Open '{filename}'")
    datafile = np.load(filename, allow_pickle=True)
    log.debug(f"Available keys {*datafile.keys(),}")
    images = datafile[path]
    if images.ndim != 6 or images.shape[-1] != 3:
        raise ValueError(f"Shape {images.shape} of '{path}' is not supported.")
    settings = json.loads(datafile["settings"][()]) if "settings" in datafile else None
    system = System.cubic(size=images.shape[1:4])
    return list(images), system, settings



from .log import log, configLog
from .property import Property, ListProperty, NumericProperty, BooleanProperty, PropertiesCollection
from .system import System
from .io import load_magnes_npz, load_lcsim_npz
from .colors import vector_to_rgb

__all__ = [
    "log",
    "configLog",
    "Property",
    "ListProperty",
    "NumericProperty",
    "BooleanProperty",
    'PropertiesCollection',
    "System",
    "load_magnes_npz",
    "load_lcsim_npz",
    "vector_to_rgb",
]
